Package: jgd
Title: JSON Graphics Device
Version: 0.1.1
Authors@R: c(
        person("Grant", "McDermott", role = c("aut", "cre"),
            email = "contact@grantmcdermott.com"),
        person("Tatsuya", "Shima", role = "aut",
            email = "ts1s1andn@gmail.com"),
        person("Dave", "Gamble", role = "cph",
            comment = "cJSON library in src/cjson/"),
        person("cJSON contributors", role = "cph",
            comment = "cJSON library in src/cjson/")
    )
Description: A graphics device that translates R plotting operations into JSON
    and streams them over a local connection to an external display
    application. The device acts as a pure recorder with no rendering
    dependencies; all rendering occurs in that application (e.g. a 'VS Code'
    extension or a web browser). Official display applications are available
    from the project homepage.
License: MIT + file LICENSE
Copyright: file inst/COPYRIGHTS
Encoding: UTF-8
NeedsCompilation: yes
Suggests: callr, ggplot2, jsonlite, processx, testthat (>= 3.0.0),
        withr
URL: https://github.com/grantmcdermott/jgd
BugReports: https://github.com/grantmcdermott/jgd/issues
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
Packaged: 2026-06-30 21:36:51 UTC; gmcd
Author: Grant McDermott [aut, cre],
  Tatsuya Shima [aut],
  Dave Gamble [cph] (cJSON library in src/cjson/),
  cJSON contributors [cph] (cJSON library in src/cjson/)
Maintainer: Grant McDermott <contact@grantmcdermott.com>
Repository: CRAN
Date/Publication: 2026-06-30 23:40:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-07-01 02:20:42 UTC; unix
Archs: jgd.so.dSYM

Package: jgd
Title: JSON Graphics Device
Version: 0.1.0
Authors@R: c(
        person("Grant", "McDermott", role = c("aut", "cre"),
            email = "contact@grantmcdermott.com"),
        person("Tatsuya", "Shima", role = "aut",
            email = "ts1s1andn@gmail.com"),
        person("Dave", "Gamble", role = "cph",
            comment = "cJSON library in src/cjson/"),
        person("cJSON contributors", role = "cph",
            comment = "cJSON library in src/cjson/")
    )
Description: A graphics device that translates R plotting operations into JSON
    and streams them over a local connection to an external display
    application. The device acts as a pure recorder with no rendering
    dependencies; all rendering occurs in that application (e.g. a 'VS Code'
    extension or a web browser). Official display applications are available
    from the project homepage.
License: MIT + file LICENSE
Copyright: file inst/COPYRIGHTS
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Suggests: callr, ggplot2, jsonlite, processx, testthat (>= 3.0.0),
        withr
URL: https://github.com/grantmcdermott/jgd
BugReports: https://github.com/grantmcdermott/jgd/issues
Config/testthat/edition: 3
Packaged: 2026-04-29 04:44:11 UTC; gmcd
Author: Grant McDermott [aut, cre],
  Tatsuya Shima [aut],
  Dave Gamble [cph] (cJSON library in src/cjson/),
  cJSON contributors [cph] (cJSON library in src/cjson/)
Maintainer: Grant McDermott <contact@grantmcdermott.com>
Repository: CRAN
Date/Publication: 2026-04-29 19:00:07 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-29 20:33:11 UTC; unix
Archs: jgd.so.dSYM

# Generated by roxygen2: do not edit by hand

export(jgd)
export(jgd_begin_group)
export(jgd_discover)
export(jgd_end_group)
export(jgd_ext)
export(jgd_frame_ext)
export(jgd_server_info)
export(with_jgd_ext)
export(with_jgd_frame_ext)
export(with_jgd_group)
importFrom(grDevices,recordGraphics)
useDynLib(jgd, .registration = TRUE)

#' JSON Graphics Device
#'
#' Opens a graphics device that streams plot operations as JSON to an external
#' renderer (e.g. VS Code extension or browser) over a Unix domain socket.
#'
#' @param width Device width in inches (default 8).
#' @param height Device height in inches (default 6).
#' @param dpi Resolution in dots per inch (default 96).
#' @param socket Socket address for the rendering server. Supports URI formats
#'   (`tcp://host:port`, `unix:///path/to/socket`) or raw Unix socket paths.
#'   If `NULL` (default), use the `jgd.socket` R option, falling back to the
#'  `JGD_SOCKET`environment variable. If `JGD_SOCKET` environment variable is
#'  also unset, the device discovers the socket via the discovery file.
#' @section Displaying plots with `jgd`:
#' It is important to note that `jgd()` does not display any plots; it only
#' streams them (i.e., converts them to a format that a JSON renderer
#' understands). To actually _display_ your plots with `jgd`, you'll need an
#' appropriate frontend. We provide two official renderers, both available for
#' install from the project repository:
#' <https://github.com/grantmcdermott/jgd>.
#'
#' - **VS Code extension.** An integrated plot pane for VS Code.
#' - **Deno server.** A standalone browser-based renderer.
#'
#' Users aren't limited to these two options. The `jgd` protocol is deliberately
#' frontend-agnostic; you can render plots with any client that reads JSONL
#' (JSON Lines). Again, please see the project repository for full documentation:
#' <https://github.com/grantmcdermott/jgd>
#' 
#' @section Debugging:
#' Set `options(jgd.debug = TRUE)` before opening the device to enable
#' frame-level diagnostic output on stderr (via `REprintf`).  This logs
#' details about `newPage`, `flush_frame`, and `poll_resize` events, which
#' is useful for diagnosing resize/replay issues.
#' @return Invisible `NULL`. The device is opened as a side effect.
#' @examplesIf interactive()
#' # Requires a running renderer (e.g., VS Code extension or Deno server).
#' # See the "Displaying plots" section above.
#' library(jgd)
#' jgd()
#' plot(1:10)
#' lines(1:10, col = "red", lwd = 3)
#' hist(rnorm(1000), col = "steelblue")
#' dev.off()
#' @export
jgd = function(
  width = 8,
  height = 6,
  dpi = 96,
  socket = NULL
) {
  if (is.null(socket)) {
    socket = getOption("jgd.socket", default = {
      jgd_socket_env = Sys.getenv("JGD_SOCKET", unset = "")
      if (nzchar(jgd_socket_env)) jgd_socket_env else NULL
    })
  } else {
    stopifnot(is.character(socket), length(socket) == 1L)
  }

  .Call(C_jgd, as.double(width), as.double(height), as.double(dpi), socket)
  invisible()
}

#' Get server information
#'
#' Returns metadata about the jgd server. When a jgd device is open
#' and connected, returns the welcome message information with
#' `connected = TRUE`. Otherwise, falls back to reading the
#' discovery file and returns information with `connected = FALSE`.
#' Returns `NULL` if no information is available from either source.
#'
#' The discovery fallback applies regardless of whether the current
#' device is a jgd device. This means `jgd_server_info()` can
#' return a non-`NULL` result even when no jgd device is open, as
#' long as a valid discovery file exists.
#'
#' @return A named list, or `NULL`.
#'
#'   When connected:
#'
#'   - **`connected`**: `TRUE`
#'   - **`server_name`**: Server name (character)
#'   - **`protocol_version`**: Protocol version (integer)
#'   - **`transport`**: Transport protocol (character)
#'   - **`server_info`**: Named character vector of key-value pairs
#'     from the server's `serverInfo` object
#'     (e.g. `c(httpUrl = "http://...")`); empty if absent
#'
#'   When not connected (discovery file fallback):
#'
#'   - **`connected`**: `FALSE`
#'   - **`server_name`**: Server name (character)
#'   - **`socket_path`**: Socket URI (character)
#'   - **`pid`**: Server process ID (integer)
#'   - **`server_info`**: Named character vector (as above)
#' @export
jgd_server_info = function() {
  path = file.path(jgd_cache_dir(), "discovery.json")
  .Call(C_jgd_server_info, path)
}

# Return the platform-specific cache directory for jgd.
# - Linux:   $XDG_CACHE_HOME/jgd or ~/.cache/jgd
# - macOS:   ~/Library/Caches/jgd
# - Windows: %LOCALAPPDATA%/jgd
jgd_cache_dir = function() {
  if (.Platform$OS.type == "windows") {
    base = Sys.getenv("LOCALAPPDATA", unset = "")
    if (nzchar(base)) return(file.path(base, "jgd"))
    base = Sys.getenv("USERPROFILE", unset = "")
    if (nzchar(base)) return(file.path(base, "AppData", "Local", "jgd"))
    stop("Cannot determine cache directory: LOCALAPPDATA and USERPROFILE are both unset")
  }
  home = normalizePath("~", mustWork = FALSE)
  if (!nzchar(home) || identical(home, "~")) {
    stop("Cannot determine cache directory: HOME is unset")
  }
  if (Sys.info()[["sysname"]] == "Darwin") {
    # On macOS, always use ~/Library/Caches/jgd to match server and tooling.
    return(file.path(home, "Library", "Caches", "jgd"))
  }
  xdg = Sys.getenv("XDG_CACHE_HOME", unset = "")
  if (nzchar(xdg)) return(file.path(xdg, "jgd"))
  file.path(home, ".cache", "jgd")
}

#' Discover a running jgd server
#'
#' Reads the jgd discovery file from the platform cache directory
#' (`~/.cache/jgd` on Linux, `~/Library/Caches/jgd` on macOS,
#' `%LOCALAPPDATA%/jgd` on Windows) and returns its contents.
#' This does not require an open jgd device — it simply reads the
#' file that a running server has written.
#'
#' @return A named list with `server_name` (character), `socket_path`
#'   (character), `pid` (integer), and `server_info` (named character
#'   vector), or `NULL` if no discovery file is found.
#' @export
jgd_discover = function() {
  path = file.path(jgd_cache_dir(), "discovery.json")
  .Call(C_jgd_discover, path)
}

#' Set extended graphics context (experimental)
#'
#' Sets extension fields that are included in every subsequent drawing
#' operation's graphics context (`gc.ext` in the JSON protocol).  This is
#' an experimental, low-level API for injecting renderer-specific properties
#' (e.g. blend modes, shadows, opacity) that go beyond R's standard graphics
#' parameters.
#'
#' @param json A single JSON string representing the extension object, or
#'   `NULL` to clear.  The string must be valid JSON (validated on the C side
#'   via cJSON); an error is raised otherwise.  Packages built on top of jgd
#'   (using e.g. jsonlite) should provide user-friendly wrappers.
#' @return Called for its side effect; returns `NULL` invisibly.
#'
#' @section Supported extension fields:
#' The Deno reference server and VS Code renderer currently support:
#'
#'
#' | Field | Canvas2D property | Example |
#' |:--|:--|:--|
#' | `blendMode` | `globalCompositeOperation` | `"multiply"` |
#' | `opacity` | `globalAlpha` | `0.5` |
#' | `shadow.blur` | `shadowBlur` | `10` |
#' | `shadow.color` | `shadowColor` | `"rgba(0,0,0,0.5)"` |
#' | `shadow.offsetX` | `shadowOffsetX` | `5` |
#' | `shadow.offsetY` | `shadowOffsetY` | `5` |
#' | `filter` | `filter` | `"blur(3px)"` |
#'
#' Custom renderers may support additional fields. Unknown fields are
#' silently ignored, so extensions are forward-compatible.
#'
#' @section Design for extension packages:
#' `jgd_ext()` is intentionally low-level — it accepts a raw JSON string.
#' Higher-level packages built on top of jgd can provide user-friendly
#' wrappers with proper argument checking, e.g.:
#'
#' ```r
#' jgd_shadow = function(blur = 0, color = "black",
#'                       offsetX = 0, offsetY = 0) {
#'   jgd_ext(jsonlite::toJSON(
#'     list(shadow = list(blur = blur, color = color,
#'                        offsetX = offsetX, offsetY = offsetY)),
#'     auto_unbox = TRUE
#'   ))
#' }
#' ```
#'
#' jgd itself has no dependency on jsonlite or any serialization library;
#' upstream packages choose their own.
#'
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @seealso [with_jgd_ext()], [jgd_frame_ext()], [jgd_begin_group()],
#'   [jgd_spec]
#' @examplesIf interactive()
#' jgd()
#'
#' # Drop shadow (scoped -- automatically cleared after the block)
#' with_jgd_ext(
#'   '{"shadow":{"blur":15,"color":"rgba(0,0,0,0.5)","offsetX":5,"offsetY":5}}',
#'   plot(1:10, pch = 19, cex = 3, col = "steelblue")
#' )
#'
#' # Semi-transparent overlay
#' with_jgd_ext('{"opacity":0.3}', {
#'   plot(1:10, pch = 19, cex = 5, col = "red")
#' })
#'
#' # Manual set/clear
#' jgd_ext('{"blendMode":"multiply"}')
#' plot(1:10)
#' jgd_ext(NULL)
#' @export
jgd_ext = function(json = NULL) {
  if (!is.null(json)) {
    stopifnot(is.character(json), length(json) == 1L)
  }
  result = .Call(C_jgd_set_ext, json)
  if (is.character(result))
    stop(result, call. = FALSE)
  invisible()
}

#' Scoped extended graphics context (experimental)
#'
#' Temporarily sets extension fields for the duration of `expr`, then clears
#' ext on exit (sets to `NULL`).
#'
#' @param json A single JSON string representing the extension object.
#'   Must be valid JSON; an error is raised otherwise.  Unlike [jgd_ext()],
#'   `NULL` is not accepted (use `jgd_ext(NULL)` to clear ext explicitly).
#' @param expr Expression to evaluate with the extension active.
#' @return The result of evaluating `expr`.
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @export
with_jgd_ext = function(json, expr) {
  stopifnot(is.character(json), length(json) == 1L)
  jgd_ext(json)
  on.exit(jgd_ext(NULL), add = TRUE)
  expr
}

#' Set frame-level extension fields (experimental)
#'
#' Sets extension fields that are included once per frame in the JSON protocol
#' (at the top level of the frame message, not per drawing operation).  This is
#' useful for frame-wide properties such as post-processing effects.
#'
#' @param json A single JSON string representing the extension object, or
#'   `NULL` or `""` to clear.  Non-empty strings must be valid JSON; an error
#'   is raised otherwise.
#' @return Called for its side effect; returns `NULL` invisibly.
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @export
jgd_frame_ext = function(json = NULL) {
  if (!is.null(json)) {
    stopifnot(is.character(json), length(json) == 1L)
  }
  result = .Call(C_jgd_set_frame_ext, json)
  if (is.character(result))
    stop(result, call. = FALSE)
  invisible()
}

#' Scoped frame-level extension fields (experimental)
#'
#' Temporarily sets frame-level extension fields for the duration of `expr`,
#' then clears them on exit.
#'
#' @param json A single JSON string representing the extension object.
#' @param expr Expression to evaluate with the frame extension active.
#' @return The result of evaluating `expr`.
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @export
with_jgd_frame_ext = function(json, expr) {
  stopifnot(is.character(json), length(json) == 1L)
  jgd_frame_ext(json)
  on.exit(jgd_frame_ext(NULL), add = TRUE)
  expr
}

#' Begin a drawing group (experimental)
#'
#' Emits a `beginGroup` operation into the drawing stream.  All subsequent
#' drawing operations until the matching [jgd_end_group()] are part of this
#' group.  The renderer may use the group's extension fields to apply effects
#' to the group as a whole.
#'
#' @param ext A single JSON string with extension fields for this group, or
#'   `NULL` for a group without extension fields.
#' @return Called for its side effect; returns `NULL` invisibly.
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @export
jgd_begin_group = function(ext = NULL) {
  if (!is.null(ext)) {
    stopifnot(is.character(ext), length(ext) == 1L)
  }
  # Unlike gc.ext (jgd_ext), which is device state embedded into every
  # drawing op by the C callbacks, group ops are standalone markers in the
  # ops stream that are not produced by any R graphics engine callback.
  # We use recordGraphics() so the .Call is recorded in R's display list
  # and replayed on resize.  recordGraphics() also executes immediately.
  recordGraphics(
    {
      result = .Call(C_jgd_begin_group, ext)
      if (is.character(result))
        stop(result, call. = FALSE)
    },
    list(ext = ext),
    env = getNamespace("jgd")
  )
  invisible()
}

#' End a drawing group (experimental)
#'
#' Emits an `endGroup` operation into the drawing stream, closing the most
#' recently opened group from [jgd_begin_group()].
#'
#' @return Called for its side effect; returns `NULL` invisibly.
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @export
jgd_end_group = function() {
  # See jgd_begin_group for why recordGraphics is used here.
  recordGraphics(
    .Call(C_jgd_end_group),
    list(),
    env = getNamespace("jgd")
  )
  # Update the snapshot after recordGraphics has added the endGroup entry
  # to the display list.  Without this, the snapshot (captured by cb_mode(0)
  # during the last drawing primitive) would not include the endGroup,
  # causing blank plots on plotIndex resize replay.
  .Call(C_jgd_update_snapshot)
  invisible()
}

#' Scoped drawing group (experimental)
#'
#' Opens a drawing group with extension fields, evaluates `expr`, then closes
#' the group on exit.
#'
#' @param ext A single JSON string with extension fields for this group, or
#'   `NULL` for a group without extension fields.
#' @param expr Expression to evaluate within the group.
#' @return The result of evaluating `expr`.
#' @section Lifecycle:
#' **Experimental.** This API may change in future versions.
#' @export
with_jgd_group = function(ext, expr) {
  jgd_begin_group(ext)
  on.exit({
    tryCatch(
      jgd_end_group(),
      error = function(e) {
        # If the device auto-closed the group at a page boundary
        # (cb_newPage resets group_depth), the cleanup endGroup would
        # fail with "endGroup without matching beginGroup".  Suppress
        # that specific error so the original error (if any) propagates.
        if (!grepl("endGroup without matching beginGroup", conditionMessage(e), fixed = TRUE))
          stop(e)
      }
    )
  }, add = TRUE)
  expr
}

# --- Unix socket tests (skip on Windows) ---

test_that("Unix: jgd() opens a device and dev.off() closes it", {
  skip_on_os("windows")
  expect_snapshot(jgd(socket = "unix:///nonexistent-jgd-test.sock"))
  expect_identical(names(dev.cur()), "jgd")
  dev.off()
})

test_that("Unix: jgd() respects JGD_SOCKET env var", {
  skip_on_os("windows")
  withr::local_envvar(JGD_SOCKET = "unix:///nonexistent-jgd-env-test.sock")
  expect_snapshot(jgd())
  dev.off()
})

test_that("Unix: jgd() respects jgd.socket option", {
  skip_on_os("windows")
  withr::local_options(jgd.socket = "unix:///nonexistent-jgd-opt-test.sock")
  withr::local_envvar(JGD_SOCKET = NA)
  expect_snapshot(jgd())
  dev.off()
})

test_that("Unix: drawing works without server connection", {
  skip_on_os("windows")
  expect_snapshot(jgd(socket = "unix:///nonexistent-jgd-draw-test.sock"))
  plot.new()
  rect(0, 0, 1, 1)
  text(0.5, 0.5, "test")
  dev.off()
})

# --- Unix socket with localhost authority (skip on Windows) ---

test_that("Unix: unix://localhost/ URI accepted", {
  skip_on_os("windows")
  expect_snapshot(jgd(socket = "unix://localhost/nonexistent-jgd-localhost-test.sock"))
  expect_identical(names(dev.cur()), "jgd")
  dev.off()
})

test_that("Unix: drawing works with unix://localhost/ URI", {
  skip_on_os("windows")
  expect_snapshot(jgd(socket = "unix://localhost/nonexistent-jgd-localhost-draw.sock"))
  plot.new()
  rect(0, 0, 1, 1)
  dev.off()
})

# --- TCP tests (cross-platform) ---

test_that("TCP: jgd() opens a device and dev.off() closes it", {
  expect_snapshot(jgd(socket = "tcp://127.0.0.1:1"))
  expect_identical(names(dev.cur()), "jgd")
  dev.off()
})

test_that("TCP: jgd() respects JGD_SOCKET env var", {
  withr::local_envvar(JGD_SOCKET = "tcp://127.0.0.1:1")
  expect_snapshot(jgd())
  dev.off()
})

test_that("TCP: jgd() respects jgd.socket option", {
  withr::local_options(jgd.socket = "tcp://127.0.0.1:1")
  withr::local_envvar(JGD_SOCKET = NA)
  expect_snapshot(jgd())
  dev.off()
})

test_that("TCP: drawing works without server connection", {
  expect_snapshot(jgd(socket = "tcp://127.0.0.1:1"))
  plot.new()
  rect(0, 0, 1, 1)
  text(0.5, 0.5, "test")
  dev.off()
})

# --- Named pipe tests (Windows only) ---

test_that("npipe: npipe:////./pipe/ URI accepted", {
  skip_if(.Platform$OS.type != "windows", "Named pipes only available on Windows")
  expect_warning(
    jgd(socket = "npipe:////./pipe/nonexistent-jgd-test"),
    "could not connect to renderer"
  )
  expect_identical(names(dev.cur()), "jgd")
  dev.off()
})

# --- Transport-independent tests ---

test_that("jgd() validates socket parameter type", {
  expect_snapshot(jgd(socket = 42), error = TRUE)
  expect_snapshot(jgd(socket = c("a", "b")), error = TRUE)
})

# Tests for beginGroup/endGroup operations in the drawing stream

# Collect all ops for the last page by combining the last complete
# (non-incremental) frame with any subsequent incremental frames.
# endGroup triggers its own incremental flush, so a single page's
# ops may span multiple frames.
last_complete_ops <- function(msgs) {
  frames <- extract_frames(msgs)
  full_idx <- which(!vapply(frames, function(f) isTRUE(f$incremental), logical(1)))
  if (length(full_idx) == 0) return(list())
  last_full <- max(full_idx)
  # Start from the complete frame's ops, then append any incremental deltas
  all_ops <- frames[[last_full]]$plot$ops
  if (last_full < length(frames)) {
    for (i in (last_full + 1):length(frames)) {
      if (isTRUE(frames[[i]]$incremental)) {
        all_ops <- c(all_ops, frames[[i]]$plot$ops)
      }
    }
  }
  all_ops
}

test_that("jgd_begin_group and jgd_end_group emit ops", {
  msgs <- with_mock_jgd({
    plot.new()
    jgd_begin_group('{"filter":"blur(5px)"}')
    rect(0, 0, 1, 1)
    jgd_end_group()
  })

  ops <- last_complete_ops(msgs)
  op_types <- vapply(ops, function(o) o$op, character(1))

  expect_true("beginGroup" %in% op_types)
  expect_true("endGroup" %in% op_types)
  expect_equal(sum(op_types == "beginGroup"), 1)
  expect_equal(sum(op_types == "endGroup"), 1)
})

test_that("beginGroup includes ext field", {
  msgs <- with_mock_jgd({
    plot.new()
    jgd_begin_group('{"filter":"blur(5px)","opacity":0.8}')
    rect(0, 0, 1, 1)
    jgd_end_group()
  })

  ops <- last_complete_ops(msgs)
  begin_ops <- Filter(function(o) identical(o$op, "beginGroup"), ops)
  expect_length(begin_ops, 1)

  ext <- begin_ops[[1]]$ext
  expect_false(is.null(ext))
  expect_equal(ext$filter, "blur(5px)")
  expect_equal(ext$opacity, 0.8)
})

test_that("beginGroup with NULL ext has no ext field", {
  msgs <- with_mock_jgd({
    plot.new()
    jgd_begin_group(NULL)
    rect(0, 0, 1, 1)
    jgd_end_group()
  })

  ops <- last_complete_ops(msgs)
  begin_ops <- Filter(function(o) identical(o$op, "beginGroup"), ops)
  expect_length(begin_ops, 1)
  expect_null(begin_ops[[1]]$ext)
})

test_that("with_jgd_group emits beginGroup and endGroup around expr", {
  msgs <- with_mock_jgd({
    plot.new()
    with_jgd_group('{"filter":"blur(3px)"}', {
      rect(0, 0, 1, 1)
    })
  })

  ops <- last_complete_ops(msgs)
  op_types <- vapply(ops, function(o) o$op, character(1))

  bg_idx <- which(op_types == "beginGroup")
  eg_idx <- which(op_types == "endGroup")
  rect_idx <- which(op_types == "rect")

  expect_equal(length(bg_idx), 1)
  expect_equal(length(eg_idx), 1)
  expect_true(length(rect_idx) >= 1)
  expect_true(bg_idx[1] < rect_idx[1])
  expect_true(eg_idx[1] > rect_idx[length(rect_idx)])
})

test_that("with_jgd_group returns expr result", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit(dev.off(), add = TRUE)

  result <- with_jgd_group('{"opacity":0.5}', 42L)
  expect_equal(result, 42L)
})

test_that("with_jgd_group emits endGroup even on error", {
  msgs <- with_mock_jgd({
    plot.new()
    tryCatch(
      with_jgd_group('{"opacity":0.5}', {
        rect(0, 0, 1, 1)
        stop("test error")
      }),
      error = function(e) NULL
    )
  })

  ops <- last_complete_ops(msgs)
  op_types <- vapply(ops, function(o) o$op, character(1))

  expect_equal(sum(op_types == "beginGroup"), 1)
  expect_equal(sum(op_types == "endGroup"), 1)
})

test_that("with_jgd_group tolerates auto-close when expr triggers new page", {
  # plot.new() inside with_jgd_group triggers cb_newPage which auto-closes
  # the group.  The on.exit jgd_end_group() should not error.
  # The first plot.new() sets up the page; the second (inside the group)
  # triggers a new page and the auto-close.
  expect_warning(
    with_mock_jgd({
      plot.new()
      with_jgd_group('{"opacity":0.5}', {
        rect(0, 0, 1, 1)
        plot.new()  # triggers cb_newPage → auto-close
      })
    }),
    "unclosed group"
  )
})

test_that("with_jgd_group re-throws non-matching endGroup errors", {
  # jgd_begin_group must succeed so we reach the on.exit cleanup.
  # Then we close the device inside expr, so jgd_end_group fails with
  # "no active graphics device" — a non-matching error that must propagate.
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit(graphics.off(), add = TRUE)

  expect_error(
    with_jgd_group('{"opacity":0.5}', {
      rect(0, 0, 1, 1)
      dev.off()  # close device so on.exit jgd_end_group fails differently
    })
  )
})

test_that("nested groups produce correct op sequence", {
  msgs <- with_mock_jgd({
    plot.new()
    jgd_begin_group('{"filter":"blur(5px)"}')
    rect(0, 0, 1, 1)
    jgd_begin_group('{"opacity":0.5}')
    rect(0.2, 0.2, 0.8, 0.8)
    jgd_end_group()
    jgd_end_group()
  })

  ops <- last_complete_ops(msgs)
  begin_ops <- Filter(function(o) identical(o$op, "beginGroup"), ops)
  end_ops <- Filter(function(o) identical(o$op, "endGroup"), ops)

  expect_length(begin_ops, 2)
  expect_length(end_ops, 2)

  expect_equal(begin_ops[[1]]$ext$filter, "blur(5px)")
  expect_equal(begin_ops[[2]]$ext$opacity, 0.5)
})

# --- Resize replay test ---

# TCP mock server that sends a resize after the first newPage frame.
start_mock_server_group_resize <- function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file <- tempfile(
    pattern = "jgd-grp-resize-port-",
    fileext = ".txt"
  )

  bg <- callr::r_bg(
    function(port_file) {
      safe_write <- function(conn, text) {
        tryCatch(
          {
            writeLines(text, conn)
            flush(conn)
          },
          error = function(e) invisible(NULL)
        )
      }

      server <- NULL
      port <- NULL
      for (i in seq_len(20)) {
        candidate <- sample(10000L:60000L, 1L)
        result <- tryCatch(
          serverSocket(candidate),
          error = function(e) NULL
        )
        if (!is.null(result)) {
          server <- result
          port <- candidate
          break
        }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn <- socketAccept(
        server,
        blocking = TRUE,
        open = "r+"
      )
      on.exit(close(conn), add = TRUE)

      messages <- list()
      new_page_count <- 0L
      resize_sent <- FALSE

      repeat {
        ready <- socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line <- tryCatch(
          readLines(conn, n = 1),
          error = function(e) character(0)
        )
        if (length(line) == 0 || !nzchar(line)) next

        msg <- tryCatch(
          jsonlite::fromJSON(
            line,
            simplifyVector = FALSE
          ),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages <- c(messages, list(msg))

        if (identical(msg$type, "metrics_request")) {
          id <- msg$id
          resp <- if (identical(msg$kind, "strWidth")) {
            str <- if (is.null(msg$str)) "" else msg$str
            list(
              type = "metrics_response",
              id = id,
              width = nchar(str) * 8.0
            )
          } else {
            list(
              type = "metrics_response",
              id = id,
              ascent = 10.0,
              descent = 3.0,
              width = 8.0
            )
          }
          safe_write(
            conn,
            jsonlite::toJSON(resp, auto_unbox = TRUE)
          )
        }

        if (identical(msg$type, "frame") &&
              isTRUE(msg$newPage)) {
          new_page_count <- new_page_count + 1L
        }

        if (new_page_count >= 1L && !resize_sent) {
          resize_sent <- TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize",
            width = 500L,
            height = 400L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port <- NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str <- readLines(
        port_file, n = 1,
        warn = FALSE
      )
      if (length(port_str) > 0 && nzchar(port_str)) {
        port <- as.integer(port_str)
        break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) {
    bg$kill()
    skip("Mock server did not start in time")
  }

  list(
    bg = bg,
    socket_url = sprintf(
      "tcp://127.0.0.1:%d",
      port
    ),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status <- bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop(
          "Mock server exited with error: ",
          bg$read_error()
        )
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("group ops survive resize replay via recordGraphics", {
  skip_on_cran()

  server <- start_mock_server_group_resize()
  withr::defer(server$cleanup())

  jgd(
    width = 4, height = 3, dpi = 72,
    socket = server$socket_url
  )

  plot.new()
  jgd_begin_group('{"filter":"blur(5px)"}')
  rect(0, 0, 1, 1)
  jgd_end_group()

  Sys.sleep(0.5)
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs <- server$collect()

  frames <- Filter(
    function(m) identical(m$type, "frame"),
    msgs
  )
  resize_frames <- Filter(
    function(f) isTRUE(f$resizeReplay),
    frames
  )
  expect_true(
    length(resize_frames) >= 1,
    info = "Should have at least 1 resize replay frame"
  )

  replay_ops <- resize_frames[[1]]$plot$ops
  op_types <- vapply(
    replay_ops,
    function(o) if (is.null(o$op)) "" else o$op,
    character(1)
  )

  expect_true(
    "beginGroup" %in% op_types,
    info = "beginGroup should survive resize replay"
  )
  expect_true(
    "endGroup" %in% op_types,
    info = "endGroup should survive resize replay"
  )

  begin_ops <- Filter(
    function(o) identical(o$op, "beginGroup"),
    replay_ops
  )
  expect_equal(
    begin_ops[[1]]$ext$filter, "blur(5px)",
    info = "Group ext should be preserved after replay"
  )
})

# --- plotIndex resize replay test (jgd-592: past plot turns blank) ---

# TCP mock server that sends a plotIndex=0 resize after receiving two
# newPage frames, then collects all messages including the replay.
start_mock_server_group_plotindex_resize <- function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file <- tempfile(
    pattern = "jgd-grp-pi-resize-port-",
    fileext = ".txt"
  )

  bg <- callr::r_bg(
    function(port_file) {
      safe_write <- function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server <- NULL
      port <- NULL
      for (i in seq_len(20)) {
        candidate <- sample(10000L:60000L, 1L)
        result <- tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server <- result; port <- candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn <- socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      messages <- list()
      new_page_count <- 0L
      resize_sent <- FALSE

      repeat {
        ready <- socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line <- tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg <- tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages <- c(messages, list(msg))

        if (identical(msg$type, "metrics_request")) {
          str_val <- if (is.null(msg$str)) "" else msg$str
          resp <- if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(str_val) * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        if (identical(msg$type, "frame") && isTRUE(msg$newPage)) {
          new_page_count <- new_page_count + 1L
        }

        # After 2 newPage frames, send plotIndex=0 resize
        if (new_page_count >= 2L && !resize_sent) {
          resize_sent <- TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize", width = 500L, height = 400L, plotIndex = 0L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port <- NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str <- readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port <- as.integer(port_str)
        break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status <- bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) { bg$kill(); skip("Mock server did not exit in time") }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("group ops survive plotIndex resize replay (jgd-592)", {
  skip_on_cran()

  server <- start_mock_server_group_plotindex_resize()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)

  # Plot 1: group with blur filter
  plot.new()
  jgd_begin_group('{"filter":"blur(5px)"}')
  rect(0, 0, 1, 1)
  jgd_end_group()

  # Plot 2: plain plot (no groups)
  plot.new()
  rect(0, 0, 0.5, 0.5, col = "red")

  # Wait for mock server to send plotIndex=0 resize
  Sys.sleep(0.5)
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs <- server$collect()

  # Find resize replay frames with plotIndex=0
  frames <- Filter(function(m) identical(m$type, "frame"), msgs)
  resize_frames <- Filter(
    function(f) isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L),
    frames
  )
  expect_true(
    length(resize_frames) >= 1,
    info = "Should have at least 1 plotIndex=0 resize replay frame"
  )

  # Collect all ops from the resize replay (may span multiple frames
  # if endGroup triggers an incremental flush)
  replay_ops <- resize_frames[[1]]$plot$ops
  # Also check for subsequent incremental frames that are part of
  # the same plotIndex replay
  replay_idx <- which(vapply(frames, function(f) {
    isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L)
  }, logical(1)))[1]
  if (replay_idx < length(frames)) {
    for (j in (replay_idx + 1):length(frames)) {
      if (isTRUE(frames[[j]]$incremental)) {
        replay_ops <- c(replay_ops, frames[[j]]$plot$ops)
      } else {
        break
      }
    }
  }

  op_types <- vapply(
    replay_ops,
    function(o) if (is.null(o$op)) "" else o$op,
    character(1)
  )

  expect_true(
    "beginGroup" %in% op_types,
    info = paste0(
      "plotIndex=0 resize replay should contain beginGroup. ",
      "Ops found: [", paste(op_types, collapse = ", "), "]"
    )
  )
  expect_true(
    "endGroup" %in% op_types,
    info = paste0(
      "plotIndex=0 resize replay should contain endGroup. ",
      "Ops found: [", paste(op_types, collapse = ", "), "]"
    )
  )

  # Group ext should be preserved
  begin_ops <- Filter(function(o) identical(o$op, "beginGroup"), replay_ops)
  expect_true(
    length(begin_ops) >= 1 && identical(begin_ops[[1]]$ext$filter, "blur(5px)"),
    info = "Group ext should be preserved in plotIndex resize replay"
  )
})

# --- Input validation tests ---

test_that("jgd_begin_group rejects invalid JSON", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(jgd_begin_group("not valid json"), "invalid JSON")
  expect_error(jgd_begin_group("{unclosed"), "invalid JSON")
})

test_that("jgd_begin_group rejects non-string input", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(jgd_begin_group(42))
  expect_error(jgd_begin_group(list(a = 1)))
})

test_that("jgd_end_group errors without matching beginGroup", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  plot.new()
  expect_error(jgd_end_group(), "endGroup without matching beginGroup")
})

# --- Auto-close tests ---

test_that("unclosed group is auto-closed at new page with warning", {
  msgs <- with_mock_jgd({
    plot.new()
    jgd_begin_group('{"opacity":0.5}')
    rect(0, 0, 1, 1)
    # No jgd_end_group() — auto-close should happen at next plot.new()
    expect_warning(plot.new(), "unclosed group")
    rect(0, 0, 0.5, 0.5)
  })

  # Collect all ops from frames before the second page's first complete frame.
  # The auto-closed endGroup may be in an incremental frame.
  frames <- extract_frames(msgs)
  full_frames <- Filter(function(f) !isTRUE(f$incremental), frames)
  expect_true(length(full_frames) >= 2)

  # All ops up to (but not including) the second complete frame belong
  # to the first page (including auto-close incremental frames).
  second_idx <- which(vapply(
    frames,
    function(f) !isTRUE(f$incremental),
    logical(1)
  ))[2]
  first_page_frames <- frames[seq_len(second_idx - 1)]
  all_ops <- unlist(
    lapply(first_page_frames, function(f) f$plot$ops),
    recursive = FALSE
  )
  op_types <- vapply(all_ops, function(o) o$op, character(1))
  begin_count <- sum(op_types == "beginGroup")
  end_count <- sum(op_types == "endGroup")
  expect_equal(begin_count, end_count,
    info = "Auto-close should balance beginGroup/endGroup")
})

test_that("unclosed group is auto-closed at device close with warning", {
  expect_warning(
    with_mock_jgd({
      plot.new()
      jgd_begin_group('{"filter":"blur(3px)"}')
      rect(0, 0, 1, 1)
      # No jgd_end_group() — auto-close should happen at dev.off()
    }),
    "unclosed group"
  )
})

test_that("multiple unclosed groups are all auto-closed", {
  msgs <- with_mock_jgd({
    plot.new()
    jgd_begin_group('{"opacity":0.5}')
    jgd_begin_group('{"filter":"blur(3px)"}')
    rect(0, 0, 1, 1)
    # 2 unclosed groups
    expect_warning(plot.new(), "2 unclosed group")
    rect(0, 0, 0.5, 0.5)
  })

  # Collect all first-page ops (including auto-close incremental frames)
  frames <- extract_frames(msgs)
  full_indices <- which(vapply(
    frames,
    function(f) !isTRUE(f$incremental),
    logical(1)
  ))
  expect_true(length(full_indices) >= 2)
  first_page_frames <- frames[seq_len(full_indices[2] - 1)]
  all_ops <- unlist(
    lapply(first_page_frames, function(f) f$plot$ops),
    recursive = FALSE
  )
  op_types <- vapply(all_ops, function(o) o$op, character(1))
  expect_equal(
    sum(op_types == "beginGroup"),
    sum(op_types == "endGroup"),
    info = "Both unclosed groups should be auto-closed"
  )
})

# --- Error tests ---

test_that("group ops errors when no jgd device is active", {
  graphics.off()
  expect_error(jgd_begin_group('{"opacity":0.5}'))
  # jgd_end_group may error with "no active graphics device",
  # "not a jgd device", or "endGroup without matching beginGroup"
  # depending on device state after graphics.off()
  expect_error(jgd_end_group())
})

# Open a jgd device without a server (connection fails with warning, but
# the device itself is usable for testing jgd_ext).
open_jgd <- function() {
  suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
}

test_that("jgd_ext accepts a JSON string", {
  open_jgd()
  on.exit(dev.off(), add = TRUE)

  expect_invisible(jgd_ext('{"blendMode":"multiply"}'))
})

test_that("jgd_ext(NULL) clears the extension", {
  open_jgd()
  on.exit(dev.off(), add = TRUE)

  jgd_ext('{"opacity":0.5}')
  expect_invisible(jgd_ext(NULL))
})

test_that("jgd_ext rejects non-string input", {
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(jgd_ext(42))
  expect_error(jgd_ext(list(a = 1)))
  expect_error(jgd_ext(TRUE))
})

test_that("jgd_ext rejects invalid JSON", {
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  # expect_snapshot cannot be used here: testthat evaluates snapshot
  # code in a context where the current graphics device is not visible,
  # so jgd_ext fails with "not a jgd device" instead of the JSON error.
  expect_error(jgd_ext("not valid json"), "invalid JSON")
  expect_error(jgd_ext("{unclosed"), "invalid JSON")
  # Empty string is treated as clearing (same as NULL)
  expect_invisible(jgd_ext(""))
})

test_that("jgd_ext errors when no jgd device is active", {
  # Close all graphics devices to ensure no jgd device is current
  graphics.off()
  expect_error(jgd_ext('{"opacity":0.5}'))
})

test_that("with_jgd_ext rejects non-string input", {
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(with_jgd_ext(42, plot(1)))
  expect_error(with_jgd_ext(NULL, plot(1)))
  expect_error(with_jgd_ext(list(a = 1), plot(1)))
})

test_that("with_jgd_ext rejects invalid JSON", {
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(with_jgd_ext("not valid json", 1), "invalid JSON")
  expect_error(with_jgd_ext("{unclosed", 1), "invalid JSON")
})

test_that("with_jgd_ext restores NULL on exit", {
  open_jgd()
  on.exit(dev.off(), add = TRUE)

  # After with_jgd_ext, ext should be cleared (NULL).
  # We can't directly read ext_json from C, but we can verify
  # that the function runs without error and returns the expr result.
  result <- with_jgd_ext('{"blendMode":"screen"}', {
    42L
  })
  expect_equal(result, 42L)
})

test_that("with_jgd_ext restores on error", {
  open_jgd()
  on.exit(dev.off(), add = TRUE)

  expect_error(
    with_jgd_ext('{"opacity":0.3}', stop("test error")),
    "test error"
  )
  # If restoration failed, subsequent jgd_ext(NULL) would crash or behave oddly
  expect_invisible(jgd_ext(NULL))
})

# Mock JSON servers for jgd testthat tests
#
# Two transport variants:
#   start_mock_server_local() — local IPC socket via processx (Unix domain socket
#                              on Unix, Windows named pipe on Windows)
#   start_mock_server_tcp()  — TCP socket (base R), works on all platforms
#
# Both run a background R process (via callr::r_bg) that:
# 1. Listens on a socket
# 2. Accepts one jgd device connection
# 3. Responds to metrics_request messages with approximate values
# 4. Collects all received JSON messages
# 5. Returns collected messages when the device sends "close"

start_mock_server_local = function(send_welcome = FALSE, transport = "unix") {
  skip_if_not_installed("callr")
  skip_if_not_installed("processx")
  skip_if_not_installed("jsonlite")

  is_windows = (.Platform$OS.type == "windows")

  if (is_windows) {
    # Windows: named pipe via processx
    pipe_name = sprintf("jgd-test-%d-%s", Sys.getpid(),
                        basename(tempfile()))
    # processx expects \\?\pipe\NAME (extended-length path prefix);
    # the C client uses \\.\pipe\NAME (device namespace) — both resolve
    # to the same kernel pipe object
    win_path = paste0("\\\\?\\pipe\\", pipe_name)
    ready_file = tempfile(pattern = "jgd-test-ready-", fileext = ".txt")
  } else {
    socket_path = tempfile(pattern = "jgd-test-", fileext = ".sock")
  }

  bg = callr::r_bg(
    function(conn_path, ready_file, send_welcome, transport) {
      `%||%` = function(x, y) if (is.null(x)) y else x
      server = processx::conn_create_unix_socket(conn_path)

      # Signal readiness: on Windows write a ready file (pipe has no filesystem
      # presence); on Unix the socket file itself signals readiness
      if (!is.null(ready_file)) {
        writeLines("ready", ready_file)
      }

      # Wait for client connection (30s timeout)
      # poll() returns "connect" (not "ready") for new connections on a
      # listening Unix socket / named pipe
      res = processx::poll(list(server), 30000)
      if (!res[[1L]] %in% c("ready", "connect")) {
        stop("No client connected within 30s (poll returned: ", res[[1L]], ")")
      }
      processx::conn_accept_unix_socket(server)

      welcome_sent = FALSE
      messages = list()
      repeat {
        res = processx::poll(list(server), 5000)
        if (!res[[1L]] %in% c("ready", "connect")) {
          next
        }

        lines = processx::conn_read_lines(server)
        for (line in lines) {
          if (!nzchar(line)) {
            next
          }
          msg = jsonlite::fromJSON(line, simplifyVector = FALSE)
          messages = c(messages, list(msg))

          # Send server_info welcome after receiving the first message
          if (send_welcome && !welcome_sent) {
            welcome = list(
              type = "server_info",
              serverName = "jgd-mock",
              protocolVersion = 1L,
              transport = transport,
              serverInfo = list(httpUrl = "http://127.0.0.1:9999/")
            )
            processx::conn_write(
              server,
              paste0(jsonlite::toJSON(welcome, auto_unbox = TRUE), "\n")
            )
            welcome_sent = TRUE
          }

          # Respond to metrics_request so tests run fast
          if (identical(msg$type, "metrics_request")) {
            resp = if (identical(msg$kind, "strWidth")) {
              list(
                type = "metrics_response",
                id = msg$id,
                width = nchar(msg$str %||% "") * 8.0
              )
            } else {
              list(
                type = "metrics_response",
                id = msg$id,
                ascent = 10.0,
                descent = 3.0,
                width = 8.0
              )
            }
            processx::conn_write(
              server,
              paste0(jsonlite::toJSON(resp, auto_unbox = TRUE), "\n")
            )
          }

          if (identical(msg$type, "close")) break
        }

        # Exit loop after close
        if (
          length(messages) > 0 &&
            identical(messages[[length(messages)]]$type, "close")
        ) {
          break
        }
      }

      close(server)
      messages
    },
    args = list(
      conn_path = if (is_windows) win_path else socket_path,
      ready_file = if (is_windows) ready_file else NULL,
      send_welcome = send_welcome,
      transport = transport
    ),
    supervise = TRUE
  )

  if (is_windows) {
    # Wait for the ready file to appear (pipe has no filesystem presence)
    for (i in seq_len(30)) {
      if (file.exists(ready_file)) break
      Sys.sleep(0.1)
    }
    if (!file.exists(ready_file)) {
      bg$kill()
      skip("Mock server pipe not ready in time")
    }
    client_uri = paste0("npipe:////./pipe/", pipe_name)
  } else {
    # Wait for the socket file to appear
    for (i in seq_len(30)) {
      if (file.exists(socket_path)) break
      Sys.sleep(0.1)
    }
    if (!file.exists(socket_path)) {
      bg$kill()
      skip("Mock server socket not created in time")
    }
    client_uri = socket_path
  }

  list(
    bg = bg,
    socket_path = client_uri,
    collect = function(timeout = 10000) {
      bg$wait(timeout)
      if (bg$get_exit_status() != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) {
        bg$kill()
      }
      if (!is_windows) {
        unlink(socket_path)
      } else {
        unlink(ready_file)
      }
    }
  )
}

# TCP mock server using base R sockets (works on all platforms including Windows)
start_mock_server_tcp = function(send_welcome = FALSE, transport = "tcp") {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file = tempfile(pattern = "jgd-tcp-port-", fileext = ".txt")

  bg = callr::r_bg(
    function(port_file, send_welcome, transport) {
      `%||%` = function(x, y) if (is.null(x)) y else x
      # Find a free port and start listening
      server = NULL
      port = NULL
      for (i in seq_len(20)) {
        candidate = sample(10000L:60000L, 1L)
        result = tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) {
          server = result
          port = candidate
          break
        }
      }
      if (is.null(port)) {
        stop("Could not find free port for TCP mock server")
      }

      on.exit(close(server), add = TRUE)

      # Signal readiness with port number
      writeLines(as.character(port), port_file)

      # Accept one client connection
      conn = socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      welcome_sent = FALSE
      messages = list()
      repeat {
        ready = socketSelect(list(conn), timeout = 5)
        if (!ready) {
          next
        }

        line = readLines(conn, n = 1)
        if (length(line) == 0 || !nzchar(line)) {
          next
        }

        msg = jsonlite::fromJSON(line, simplifyVector = FALSE)
        messages = c(messages, list(msg))

        # Send server_info welcome after receiving the first message
        if (send_welcome && !welcome_sent) {
          welcome = list(
            type = "server_info",
            serverName = "jgd-mock",
            protocolVersion = 1L,
            transport = transport,
            serverInfo = list(httpUrl = "http://127.0.0.1:9999/")
          )
          writeLines(jsonlite::toJSON(welcome, auto_unbox = TRUE), conn)
          flush(conn)
          welcome_sent = TRUE
        }

        # Respond to metrics_request so tests run fast
        if (identical(msg$type, "metrics_request")) {
          resp = if (identical(msg$kind, "strWidth")) {
            list(
              type = "metrics_response",
              id = msg$id,
              width = nchar(msg$str %||% "") * 8.0
            )
          } else {
            list(
              type = "metrics_response",
              id = msg$id,
              ascent = 10.0,
              descent = 3.0,
              width = 8.0
            )
          }
          writeLines(jsonlite::toJSON(resp, auto_unbox = TRUE), conn)
          flush(conn)
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file, send_welcome = send_welcome, transport = transport),
    supervise = TRUE
  )

  # Wait for the port file to appear (server is listening)
  port = NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str = readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port = as.integer(port_str)
        break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) {
    bg$kill()
    skip("Mock TCP server did not start in time")
  }

  list(
    bg = bg,
    port = port,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 10000) {
      bg$wait(timeout)
      if (bg$get_exit_status() != 0) {
        stop("Mock TCP server exited with error: ", bg$read_error())
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) {
        bg$kill()
      }
      unlink(port_file)
    }
  )
}

# Convenience: open jgd device connected to mock server, run expr, close,
# and return all captured JSON messages.
with_mock_jgd = function(
  expr,
  width = 4,
  height = 3,
  dpi = 72,
  transport = c("unix", "tcp"),
  send_welcome = FALSE
) {
  transport = match.arg(transport)
  if (transport == "tcp") {
    server = start_mock_server_tcp(send_welcome = send_welcome)
    socket_addr = server$socket_url
  } else {
    server = start_mock_server_local(send_welcome = send_welcome)
    socket_addr = server$socket_path
  }
  withr::defer(server$cleanup())

  jgd(width = width, height = height, dpi = dpi, socket = socket_addr)
  force(expr)
  dev.off()

  server$collect()
}

# Extract frame messages from a list of collected messages
extract_frames = function(msgs) {
  Filter(function(m) identical(m$type, "frame"), msgs)
}

# Extract all ops across all frames
extract_ops = function(msgs) {
  frames = extract_frames(msgs)
  unlist(lapply(frames, function(f) f$plot$ops), recursive = FALSE)
}

# Extract ops of a specific type
extract_ops_by_type = function(msgs, op_type) {
  ops = extract_ops(msgs)
  Filter(function(o) identical(o$op, op_type), ops)
}

# Write a discovery file to a temporary directory and mock jgd_cache_dir()
# to point there. Returns the temporary cache dir path.
write_test_discovery = function(json_content, envir = parent.frame()) {
  cache_dir = withr::local_tempdir("jgd-cache-", .local_envir = envir)
  jgd_dir = file.path(cache_dir, "jgd")
  dir.create(jgd_dir, recursive = TRUE)
  writeLines(json_content, file.path(jgd_dir, "discovery.json"))
  testthat::local_mocked_bindings(
    jgd_cache_dir = function() file.path(cache_dir, "jgd"),
    .package = "jgd",
    .env = envir
  )
  cache_dir
}

# Mock jgd_cache_dir() to return an empty directory (no discovery file).
set_empty_discovery_env = function(envir = parent.frame()) {
  empty_dir = withr::local_tempdir("jgd-empty-", .local_envir = envir)
  testthat::local_mocked_bindings(
    jgd_cache_dir = function() file.path(empty_dir, "jgd"),
    .package = "jgd",
    .env = envir
  )
}

# Snapshot an R value as pretty-printed JSON matching jgd wire format
expect_json_snapshot = function(x) {
  json = jsonlite::toJSON(x, auto_unbox = TRUE, pretty = TRUE)
  expect_snapshot(cat(json))
}

# Regression test for lines() disappearing on plotIndex resize replay.
#
# When plot(1:10); lines(1:10, col="red") is followed by hist(), navigating
# back to plot 1 and resizing should preserve the lines() polyline ops.
#
# This test is in a separate file from test-resize-plotindex-content.R to
# avoid grid state contamination from ggplot2 tests that run in the same
# file (ggplot2's grid state persists within a devtools::test() session and
# interferes with GEplaySnapshot for base graphics).

# TCP mock server that sends a plotIndex=0 resize after receiving two
# newPage frames, then collects all messages including the replay.
start_mock_server_plotindex_lines = function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file = tempfile(pattern = "jgd-pi-lines-port-", fileext = ".txt")

  bg = callr::r_bg(
    function(port_file) {
      safe_write = function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server = NULL; port = NULL
      for (i in seq_len(20)) {
        candidate = sample(10000L:60000L, 1L)
        result = tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server = result; port = candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn = socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      messages = list()
      new_page_count = 0L
      resize_sent = FALSE

      repeat {
        ready = socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line = tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg = tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages = c(messages, list(msg))

        # Respond to metrics_request
        if (identical(msg$type, "metrics_request")) {
          resp = if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(if (is.null(msg$str)) "" else msg$str) * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        # Count newPage frames
        if (identical(msg$type, "frame") && isTRUE(msg$newPage)) {
          new_page_count = new_page_count + 1L
        }

        # After 2 newPage frames, send a plotIndex=0 resize
        if (new_page_count >= 2L && !resize_sent) {
          resize_sent = TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize", width = 500L, height = 400L, plotIndex = 0L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port = NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str = readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port = as.integer(port_str); break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status = bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("plotIndex resize replay preserves lines() added after plot()", {
  skip_on_cran()

  server = start_mock_server_plotindex_lines()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)

  # Plot 1: plot + lines (two drawing calls on the same page)
  plot(1:10)
  lines(1:10, col = "red", lwd = 3)

  # Plot 2: hist (new page)
  hist(rnorm(1000), col = "steelblue")

  # Wait for the mock server to send the plotIndex=0 resize
  Sys.sleep(1.5)

  # Process the plotIndex=0 resize
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs = server$collect()

  # Find resize replay frames with plotIndex=0
  frames = Filter(function(m) identical(m$type, "frame"), msgs)
  resize_frames = Filter(
    function(f) isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L),
    frames
  )
  expect_true(length(resize_frames) >= 1,
    info = "Should have at least 1 plotIndex=0 resize replay frame")

  # Collect all ops from the replay (including incremental frames)
  replay_ops = resize_frames[[1]]$plot$ops
  replay_idx = which(vapply(frames, function(f) {
    isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L)
  }, logical(1)))[1]
  if (replay_idx < length(frames)) {
    for (j in (replay_idx + 1):length(frames)) {
      if (isTRUE(frames[[j]]$incremental)) {
        replay_ops = c(replay_ops, frames[[j]]$plot$ops)
      } else {
        break
      }
    }
  }

  op_types = vapply(
    replay_ops,
    function(o) if (is.null(o$op)) "" else o$op,
    character(1)
  )

  # The replay must contain polyline ops from lines()
  expect_true("polyline" %in% op_types,
    info = paste0(
      "plotIndex=0 resize replay should contain polyline ops from lines(). ",
      "Ops found: [", paste(unique(op_types), collapse = ", "), "]"))

  # Verify the polyline has the red color from lines()
  polyline_ops = Filter(function(o) identical(o$op, "polyline"), replay_ops)
  expect_true(length(polyline_ops) >= 1,
    info = "Should have at least 1 polyline op")

  # Check that at least one polyline has a red-ish color
  # jgd may serialize as "#FF0000FF" or "rgba(255,0,0,1)"
  has_red = any(vapply(polyline_ops, function(o) {
    col = o$gc$col
    if (is.null(col)) return(FALSE)
    grepl("^#[Ff][Ff]0000", col) || grepl("rgba\\(255,\\s*0,\\s*0", col)
  }, logical(1)))
  expect_true(has_red,
    info = paste0(
      "plotIndex=0 replay should have a red polyline from lines(col='red'). ",
      "polyline colors: [",
      paste(vapply(polyline_ops, function(o) {
        if (is.null(o$gc$col)) "NULL" else o$gc$col
      }, character(1)), collapse = ", "), "]"))
})

test_that("jgd_server_info() returns connected=TRUE with welcome metadata", {
  skip_on_os("windows")

  server = start_mock_server_local(send_welcome = TRUE)
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_path)

  info = jgd_server_info()
  expect_type(info, "list")
  expect_true(info$connected)
  expect_identical(info$server_name, "jgd-mock")
  expect_identical(info$protocol_version, 1L)
  expect_identical(info$transport, "unix")
  expect_type(info$server_info, "character")
  expect_identical(info$server_info[["httpUrl"]], "http://127.0.0.1:9999/")

  dev.off()
  server$collect()
})

test_that("jgd_server_info() returns connected=TRUE over TCP", {
  server = start_mock_server_tcp(send_welcome = TRUE)
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)

  info = jgd_server_info()
  expect_type(info, "list")
  expect_true(info$connected)
  expect_identical(info$server_name, "jgd-mock")
  expect_identical(info$protocol_version, 1L)
  expect_identical(info$transport, "tcp")
  expect_identical(info$server_info[["httpUrl"]], "http://127.0.0.1:9999/")

  dev.off()
  server$collect()
})

test_that("jgd_server_info() falls back to discovery file when no welcome", {
  skip_on_os("windows")

  server = start_mock_server_local(send_welcome = FALSE)
  withr::defer(server$cleanup())

  # Write a discovery file
  write_test_discovery(
    '{"serverName":"disc-server","socketPath":"unix:///tmp/test.sock","pid":999}'
  )

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_path)

  info = jgd_server_info()
  expect_type(info, "list")
  expect_false(info$connected)
  expect_identical(info$server_name, "disc-server")
  expect_identical(info$socket_path, "unix:///tmp/test.sock")
  expect_identical(info$pid, 999L)

  dev.off()
  server$collect()
})

test_that("jgd_server_info() returns NULL for non-jgd device with no discovery", {
  set_empty_discovery_env()

  pdf(tempfile(fileext = ".pdf"))
  withr::defer(dev.off())

  expect_null(jgd_server_info())
})

test_that("jgd_server_info() returns discovery fallback when not connected", {
  skip_on_os("windows")

  write_test_discovery(
    '{"serverName":"fallback","socketPath":"unix:///tmp/fb.sock","pid":42}'
  )

  expect_warning(
    jgd(socket = "unix:///nonexistent-jgd-info-test.sock"),
    "could not connect"
  )

  info = jgd_server_info()
  expect_type(info, "list")
  expect_false(info$connected)
  expect_identical(info$server_name, "fallback")

  dev.off()
})

# Tests that malformed JSON from the server does not crash the device.
#
# The device receives JSON messages from the rendering server (resize,
# metrics_response).  If the server sends garbage, the device must
# silently ignore it and continue producing correct output.

# TCP mock server that injects malformed JSON at two points:
# 1. Immediately after connection (drained by check_incoming on newPage)
# 2. Before each metrics_response (drained by recv_metrics_response)
start_mock_server_malformed = function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file = tempfile(pattern = "jgd-malformed-port-", fileext = ".txt")

  bg = callr::r_bg(
    function(port_file) {
      `%||%` = function(x, y) if (is.null(x)) y else x
      safe_write = function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server = NULL; port = NULL
      for (i in seq_len(20)) {
        candidate = sample(10000L:60000L, 1L)
        result = tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server = result; port = candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn = socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      # Inject malformed JSON immediately after connection
      safe_write(conn, c(
        "{not valid json at all",
        "totally not json",
        '{"type": "resize", "width": "not_a_number", "height": "bad"}',
        '{"type": "resize"}',
        ""
      ))

      messages = list()
      repeat {
        ready = socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line = tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg = tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages = c(messages, list(msg))

        if (identical(msg$type, "metrics_request")) {
          # Inject one malformed line before the valid response
          safe_write(conn, "{broken json")

          resp = if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(msg$str %||% "") * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port = NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str = readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port = as.integer(port_str); break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status = bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("device survives malformed JSON and produces correct output", {
  skip_on_cran()

  server = start_mock_server_malformed()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)
  plot.new()
  rect(0, 0, 1, 1, col = "red", border = "blue", lwd = 2)
  text(0.5, 0.5, "hi")
  dev.off()

  msgs = server$collect()
  frames = Filter(function(m) identical(m$type, "frame"), msgs)

  # Device produced at least one frame
  expect_true(length(frames) >= 1)

  # Frame metadata is intact
  frame = frames[[1]]
  expect_equal(frame$plot$version, 1)
  expect_match(frame$plot$sessionId, "^r-[0-9]+-[0-9]+$")

  # Ops are present and correct
  ops = unlist(lapply(frames, function(f) f$plot$ops), recursive = FALSE)
  op_types = vapply(ops, function(o) o$op, character(1))
  expect_true("rect" %in% op_types)
  expect_true("text" %in% op_types)
})

test_that("malformed JSON: device metadata matches snapshot", {
  skip_on_cran()

  server = start_mock_server_malformed()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)
  plot.new()
  rect(0, 0, 1, 1, col = "red", border = "blue", lwd = 2)
  dev.off()

  msgs = server$collect()
  frame = Filter(function(m) identical(m$type, "frame"), msgs)[[1]]

  expect_json_snapshot(list(
    version = frame$plot$version,
    device = frame$plot$device
  ))
})

test_that("malformed JSON: rect gc fields match snapshot", {
  skip_on_cran()

  server = start_mock_server_malformed()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)
  plot.new()
  rect(0, 0, 1, 1, col = "red", border = "blue", lwd = 2)
  dev.off()

  msgs = server$collect()
  rect_ops = extract_ops_by_type(msgs, "rect")
  expect_json_snapshot(rect_ops[[1]]$gc)
})

# Tests for socket discovery behavior.
#
# When the user provides an explicit socket= address, the device must
# never silently fall back to the discovery file.  This prevents
# confusing connections to unintended servers when the explicit path
# is stale.

test_that("explicit socket= does not fall back to discovery file", {
  # Start a TCP mock server that we can detect connections to.
  server = start_mock_server_tcp()
  withr::defer(server$cleanup())

  # Write a discovery file pointing to the mock server.
  write_test_discovery(
    sprintf('{"serverName":"test","socketPath":"%s","pid":1}', server$socket_url)
  )
  withr::local_envvar(JGD_SOCKET = NA)
  withr::local_options(jgd.socket = NULL)

  # Open jgd with an explicit bogus socket (nothing listens on port 1).
  # The device should warn but NOT fall back to the discovery file.
  expect_warning(
    jgd(socket = "tcp://127.0.0.1:1"),
    "could not connect"
  )
  plot.new()
  rect(0, 0, 1, 1)
  dev.off()

  # If fallback had occurred, the mock server would have accepted a
  # connection, received the "close" message from dev.off(), and exited.
  # Wait long enough for that to happen, then verify it is still alive
  # (blocked on socketAccept = no connection was made).
  server$bg$wait(2000)
  expect_true(server$bg$is_alive())
})

test_that("jgd_discover reads all discovery file fields", {
  write_test_discovery(
    '{"serverName":"test-server","socketPath":"unix:///tmp/test.sock","pid":12345,"serverInfo":{"httpUrl":"http://127.0.0.1:8080/"}}'
  )

  info = jgd_discover()
  expect_type(info, "list")
  expect_equal(info$server_name, "test-server")
  expect_equal(info$socket_path, "unix:///tmp/test.sock")
  expect_equal(info$pid, 12345L)
  expect_equal(info$server_info, c(httpUrl = "http://127.0.0.1:8080/"))
})

test_that("jgd_discover returns NULL when no discovery file", {
  set_empty_discovery_env()
  expect_null(jgd_discover())
})

test_that("jgd_discover returns NULL for invalid discovery file", {
  write_test_discovery('{"socketPath":"unix:///tmp/test.sock"}')

  # Missing serverName → invalid
  expect_null(jgd_discover())
})

test_that("jgd_discover handles missing serverInfo gracefully", {
  write_test_discovery(
    '{"serverName":"test","socketPath":"unix:///tmp/test.sock","pid":1}'
  )

  info = jgd_discover()
  expect_type(info, "list")
  expect_equal(info$server_name, "test")
  expect_length(info$server_info, 0)
})

# Tests that recv_metrics_response preserves the first buffered plotIndex
# resize when multiple arrive during a single metrics exchange.
#
# Scenario: the rendering server sends two plotIndex resize messages
# BEFORE the metrics_response.  recv_metrics_response must buffer only
# the first (FIFO matches the server's pendingResizes queue) and skip
# the second.
#
# Verification: after poll_resize_impl processes the buffered resize,
# the device dimensions should match the FIRST plotIndex resize, not
# the second.  This confirms FIFO ordering was preserved.
#
# Without the fix, recv_metrics_response overwrites the buffer with the
# second plotIndex → wrong dimensions applied → server queue desync.

# TCP mock server that injects two plotIndex resize messages before the
# first metrics_response.  Subsequent metrics_requests are answered normally.
start_mock_server_dual_plotindex = function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file = tempfile(pattern = "jgd-dual-pi-port-", fileext = ".txt")

  bg = callr::r_bg(
    function(port_file) {
      `%||%` = function(x, y) if (is.null(x)) y else x
      safe_write = function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server = NULL; port = NULL
      for (i in seq_len(20)) {
        candidate = sample(10000L:60000L, 1L)
        result = tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server = result; port = candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn = socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      injected = FALSE
      messages = list()

      repeat {
        ready = socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line = tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg = tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages = c(messages, list(msg))

        if (identical(msg$type, "metrics_request")) {
          if (!injected) {
            # First metrics_request: inject two plotIndex resizes before
            # the metrics_response.  recv_metrics_response should buffer
            # only the first (plotIndex=0 at 500x400).
            safe_write(conn, jsonlite::toJSON(list(
              type = "resize", width = 500L, height = 400L, plotIndex = 0L
            ), auto_unbox = TRUE))
            safe_write(conn, jsonlite::toJSON(list(
              type = "resize", width = 600L, height = 450L, plotIndex = 1L
            ), auto_unbox = TRUE))
            injected = TRUE
          }

          resp = if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(msg$str %||% "") * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port = NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str = readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port = as.integer(port_str); break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status = bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("recv_metrics_response preserves first plotIndex resize, skips second", {
  skip_on_cran()

  server = start_mock_server_dual_plotindex()
  withr::defer(server$cleanup())

  dpi = 72
  jgd(width = 4, height = 3, dpi = dpi, socket = server$socket_url)

  # Plot 1: creates a snapshot for historical resize
  plot.new()
  rect(0, 0, 1, 1, col = "red")

  # Plot 2: text() triggers metrics_request → mock server injects
  # plotIndex=0 resize at 500x400 and plotIndex=1 resize at 600x450
  # before the metrics_response.
  # recv_metrics_response should buffer only plotIndex=0 (the first).
  plot.new()
  text(0.5, 0.5, "X")

  # Force resize processing — poll_resize_impl drains the buffer and
  # applies the buffered dimensions to the device.
  result = .Call(jgd:::C_jgd_poll_resize)
  expect_true(result, info = "poll_resize should process the buffered resize")

  # Verify device dimensions match the FIRST plotIndex resize (500x400),
  # not the second (600x450).  poll_resize_impl applies pending_w/h to
  # the device before the snapshot replay, so dev.size() reflects which
  # resize was buffered.
  size_px = dev.size("px")
  expect_equal(size_px[1], 500,
    info = "Device width should be 500 (from first plotIndex resize)")
  expect_equal(size_px[2], 400,
    info = "Device height should be 400 (from first plotIndex resize)")

  # Second poll should be no-op — only one resize was buffered
  result2 = .Call(jgd:::C_jgd_poll_resize)
  expect_false(result2,
    info = "Second poll_resize should be no-op (second plotIndex was skipped)")

  dev.off()
  server$collect()
})

# Tests for frame-level ext field

test_that("jgd_frame_ext adds ext to frame message", {
  msgs <- with_mock_jgd({
    jgd_frame_ext('{"postEffects":[{"type":"blur","radius":5}]}')
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames <- extract_frames(msgs)
  expect_true(length(frames) >= 1)

  frame <- frames[[1]]
  expect_false(is.null(frame$ext))
  expect_length(frame$ext$postEffects, 1)
  expect_equal(frame$ext$postEffects[[1]]$type, "blur")
  expect_equal(frame$ext$postEffects[[1]]$radius, 5)
})

test_that("frame ext is absent when not set", {
  msgs <- with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames <- extract_frames(msgs)
  expect_true(length(frames) >= 1)
  expect_null(frames[[1]]$ext)
})

test_that("jgd_frame_ext(NULL) clears frame ext", {
  msgs <- with_mock_jgd({
    jgd_frame_ext('{"postEffects":[{"type":"blur"}]}')
    plot.new()
    rect(0, 0, 1, 1)
    # Clear and draw second plot
    jgd_frame_ext(NULL)
    plot.new()
    rect(0, 0, 0.5, 0.5)
  })

  frames <- extract_frames(msgs)
  expect_true(length(frames) >= 2)

  # First frame should have ext
  expect_false(is.null(frames[[1]]$ext))

  # Find the second full (non-incremental) frame
  full_frames <- Filter(function(f) !isTRUE(f$incremental), frames)
  expect_true(length(full_frames) >= 2)
  expect_null(full_frames[[2]]$ext)
})

test_that("with_jgd_frame_ext scopes frame ext", {
  msgs <- with_mock_jgd({
    with_jgd_frame_ext('{"postEffects":[{"type":"glow"}]}', {
      plot.new()
      rect(0, 0, 1, 1)
    })
    # After scope, draw another plot — frame ext should be cleared
    plot.new()
    rect(0, 0, 0.5, 0.5)
  })

  frames <- extract_frames(msgs)
  full_frames <- Filter(function(f) !isTRUE(f$incremental), frames)
  expect_true(length(full_frames) >= 2)

  expect_false(is.null(full_frames[[1]]$ext))
  expect_null(full_frames[[2]]$ext)
})

test_that("with_jgd_frame_ext returns expr result", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit(dev.off(), add = TRUE)

  result <- with_jgd_frame_ext('{"postEffects":[]}', 42L)
  expect_equal(result, 42L)
})

test_that("jgd_frame_ext rejects invalid JSON", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(jgd_frame_ext("not valid json"), "invalid JSON")
  expect_error(jgd_frame_ext("{unclosed"), "invalid JSON")
  # Empty string clears (same as NULL)
  expect_invisible(jgd_frame_ext(""))
})

test_that("jgd_frame_ext rejects non-string input", {
  open_jgd <- function() suppressWarnings(jgd(socket = "tcp://127.0.0.1:1"))
  open_jgd()
  on.exit({ graphics.off() }, add = TRUE)

  expect_error(jgd_frame_ext(42))
  expect_error(jgd_frame_ext(list(a = 1)))
})

test_that("jgd_frame_ext errors when no jgd device is active", {
  graphics.off()
  expect_error(jgd_frame_ext('{"postEffects":[]}'))
})

# --- Resize replay tests ---

# TCP mock server that sends a resize after the first newPage frame,
# then collects messages including the replay frame.
start_mock_server_frame_ext_resize <- function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file <- tempfile(pattern = "jgd-fext-resize-port-", fileext = ".txt")

  bg <- callr::r_bg(
    function(port_file) {
      `%||%` <- function(x, y) if (is.null(x)) y else x
      safe_write <- function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server <- NULL; port <- NULL
      for (i in seq_len(20)) {
        candidate <- sample(10000L:60000L, 1L)
        result <- tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server <- result; port <- candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn <- socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      messages <- list()
      new_page_count <- 0L
      resize_sent <- FALSE

      repeat {
        ready <- socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line <- tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg <- tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages <- c(messages, list(msg))

        if (identical(msg$type, "metrics_request")) {
          resp <- if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(msg$str %||% "") * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        if (identical(msg$type, "frame") && isTRUE(msg$newPage)) {
          new_page_count <- new_page_count + 1L
        }

        # After first newPage frame, send a normal resize
        if (new_page_count >= 1L && !resize_sent) {
          resize_sent <- TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize", width = 500L, height = 400L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port <- NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str <- readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port <- as.integer(port_str); break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status <- bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("frame ext survives resize replay", {
  skip_on_cran()

  server <- start_mock_server_frame_ext_resize()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)
  jgd_frame_ext('{"postEffects":[{"type":"blur","radius":5}]}')

  plot(1:3)

  # Wait for the mock server to send the resize.
  # Windows i386 CI can be slow; use a longer sleep.
  Sys.sleep(1.5)

  # Process the resize
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs <- server$collect()

  frames <- Filter(function(m) identical(m$type, "frame"), msgs)

  # Should have the original frame and a resize replay frame
  resize_frames <- Filter(function(f) isTRUE(f$resizeReplay), frames)
  expect_true(length(resize_frames) >= 1,
    info = "Should have at least 1 resize replay frame")

  # The resize replay frame must preserve the frame-level ext
  replay <- resize_frames[[1]]
  expect_false(is.null(replay$ext),
    info = "Resize replay frame should preserve frame-level ext")
  expect_equal(replay$ext$postEffects[[1]]$type, "blur")
  expect_equal(replay$ext$postEffects[[1]]$radius, 5)
})

# TCP mock server that sends a plotIndex=0 resize after receiving two
# newPage frames, to test historical plot frame ext preservation.
start_mock_server_frame_ext_plotindex <- function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file <- tempfile(
    pattern = "jgd-fext-pi-port-",
    fileext = ".txt"
  )

  bg <- callr::r_bg(
    function(port_file) {
      safe_write <- function(conn, text) {
        tryCatch(
          {
            writeLines(text, conn)
            flush(conn)
          },
          error = function(e) invisible(NULL)
        )
      }

      server <- NULL
      port <- NULL
      for (i in seq_len(20)) {
        candidate <- sample(10000L:60000L, 1L)
        result <- tryCatch(
          serverSocket(candidate),
          error = function(e) NULL
        )
        if (!is.null(result)) {
          server <- result
          port <- candidate
          break
        }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn <- socketAccept(
        server,
        blocking = TRUE,
        open = "r+"
      )
      on.exit(close(conn), add = TRUE)

      messages <- list()
      new_page_count <- 0L
      resize_sent <- FALSE

      repeat {
        ready <- socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line <- tryCatch(
          readLines(conn, n = 1),
          error = function(e) character(0)
        )
        if (length(line) == 0 || !nzchar(line)) next

        msg <- tryCatch(
          jsonlite::fromJSON(
            line,
            simplifyVector = FALSE
          ),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages <- c(messages, list(msg))

        if (identical(msg$type, "metrics_request")) {
          id <- msg$id
          resp <- if (identical(msg$kind, "strWidth")) {
            str <- if (is.null(msg$str)) "" else msg$str
            list(
              type = "metrics_response",
              id = id,
              width = nchar(str) * 8.0
            )
          } else {
            list(
              type = "metrics_response",
              id = id,
              ascent = 10.0,
              descent = 3.0,
              width = 8.0
            )
          }
          safe_write(
            conn,
            jsonlite::toJSON(resp, auto_unbox = TRUE)
          )
        }

        if (identical(msg$type, "frame") &&
              isTRUE(msg$newPage)) {
          new_page_count <- new_page_count + 1L
        }

        # After 2 newPage frames, send plotIndex=0 resize
        if (new_page_count >= 2L && !resize_sent) {
          resize_sent <- TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize",
            width = 500L,
            height = 400L,
            plotIndex = 0L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port <- NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str <- readLines(
        port_file, n = 1,
        warn = FALSE
      )
      if (length(port_str) > 0 && nzchar(port_str)) {
        port <- as.integer(port_str)
        break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) {
    bg$kill()
    skip("Mock server did not start in time")
  }

  list(
    bg = bg,
    socket_url = sprintf(
      "tcp://127.0.0.1:%d",
      port
    ),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status <- bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop(
          "Mock server exited with error: ",
          bg$read_error()
        )
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("frame ext survives plotIndex resize replay", {
  skip_on_cran()

  server <- start_mock_server_frame_ext_plotindex()
  withr::defer(server$cleanup())

  jgd(
    width = 4, height = 3, dpi = 72,
    socket = server$socket_url
  )

  # Plot 1: with frame ext "blur"
  jgd_frame_ext(
    '{"postEffects":[{"type":"blur","radius":5}]}'
  )
  plot(1:3)

  # Plot 2: different frame ext "glow"
  jgd_frame_ext(
    '{"postEffects":[{"type":"glow"}]}'
  )
  plot(4:6)

  # Wait for mock server to send plotIndex=0 resize
  Sys.sleep(0.5)
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs <- server$collect()

  frames <- Filter(
    function(m) identical(m$type, "frame"),
    msgs
  )
  resize_frames <- Filter(
    function(f) {
      isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L)
    },
    frames
  )
  expect_true(
    length(resize_frames) >= 1,
    info = "Should have a plotIndex=0 resize replay frame"
  )

  replay <- resize_frames[[1]]
  expect_false(
    is.null(replay$ext),
    info = paste(
      "plotIndex=0 replay should preserve",
      "plot 1's frame ext"
    )
  )
  expect_equal(
    replay$ext$postEffects[[1]]$type, "blur",
    info = "Should be plot 1's ext (blur), not plot 2's"
  )
  expect_equal(
    replay$ext$postEffects[[1]]$radius, 5
  )
})

test_that("plotIndex replay preserves ext when later plot clears it", {
  skip_on_cran()

  server <- start_mock_server_frame_ext_plotindex()
  withr::defer(server$cleanup())

  jgd(
    width = 4, height = 3, dpi = 72,
    socket = server$socket_url
  )

  # Plot 1: with frame ext "blur"
  jgd_frame_ext(
    '{"postEffects":[{"type":"blur","radius":3}]}'
  )
  plot(1:3)

  # Plot 2: clear frame ext
  jgd_frame_ext(NULL)
  plot(4:6)

  # Wait for mock server to send plotIndex=0 resize
  Sys.sleep(0.5)
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs <- server$collect()

  frames <- Filter(
    function(m) identical(m$type, "frame"),
    msgs
  )
  resize_frames <- Filter(
    function(f) {
      isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L)
    },
    frames
  )
  expect_true(
    length(resize_frames) >= 1,
    info = "Should have a plotIndex=0 resize replay frame"
  )

  replay <- resize_frames[[1]]
  expect_false(
    is.null(replay$ext),
    info = "plotIndex=0 replay should preserve plot 1's ext even when plot 2 cleared it"
  )
  expect_equal(
    replay$ext$postEffects[[1]]$type, "blur",
    info = "Should be plot 1's ext (blur)"
  )
  expect_equal(
    replay$ext$postEffects[[1]]$radius, 3
  )
})

# --- Independence test ---

test_that("frame ext is independent of gc ext", {
  msgs <- with_mock_jgd({
    jgd_frame_ext('{"postEffects":[{"type":"blur"}]}')
    jgd_ext('{"blendMode":"multiply"}')
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames <- extract_frames(msgs)
  frame <- frames[[1]]

  # Frame-level ext
  expect_false(is.null(frame$ext))
  expect_equal(frame$ext$postEffects[[1]]$type, "blur")

  # gc-level ext on drawing ops
  rect_ops <- extract_ops_by_type(msgs, "rect")
  expect_true(length(rect_ops) >= 1)
  gc_ext <- rect_ops[[1]]$gc$ext
  expect_false(is.null(gc_ext))
  expect_equal(gc_ext$blendMode, "multiply")
})

test_that("frame ext glow + gc.ext opacity produces correct frame data", {
  # Regression test for jgd-8qz: glow post-effect with gc.ext opacity
  # rendered blank in the browser due to additive blending on white bg.
  # This test verifies the R-side frame data is correctly formed.
  msgs <- with_mock_jgd({
    jgd_frame_ext('{"postEffects":[{"type":"glow"}]}')
    jgd_ext('{"opacity":0.5}')
    plot.new()
    rect(0.1, 0.1, 0.9, 0.9, col = "red")
    jgd_ext(NULL)
    jgd_frame_ext(NULL)
  })

  frames <- extract_frames(msgs)
  expect_true(length(frames) >= 1)

  frame <- frames[[1]]
  # Frame should have glow postEffect
  expect_false(is.null(frame$ext))
  expect_equal(frame$ext$postEffects[[1]]$type, "glow")

  # Drawing ops should have gc.ext with opacity
  rect_ops <- Filter(
    function(o) identical(o$op, "rect"),
    frame$plot$ops
  )
  expect_true(length(rect_ops) >= 1)
  expect_equal(rect_ops[[1]]$gc$ext$opacity, 0.5)
})

# Tests that resize replay frames carry the correct plotNumber.
#
# Bug: cb_newPage increments page_count even during replays (st->replaying),
# so the plotNumber in the resize replay frame (page_count - 1) is one
# higher than the original plot's plotNumber.  The browser-side
# replaceLatest guard compares the replay's plotNumber against the stored
# _rIndex and rejects the update when they don't match, causing the plot
# to not be re-rendered at the new size.
#
# This test creates two plots, sends a normal resize, and verifies that
# the resize replay frame's plotNumber matches the original plot 2's
# plotNumber (not page_count - 1 after the replay incremented it).

# TCP mock server that sends a resize message after receiving two newPage
# frames, then collects the resize replay frame.
start_mock_server_resize_plotnumber = function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file = tempfile(pattern = "jgd-resize-pn-port-", fileext = ".txt")

  bg = callr::r_bg(
    function(port_file) {
      `%||%` = function(x, y) if (is.null(x)) y else x
      safe_write = function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server = NULL; port = NULL
      for (i in seq_len(20)) {
        candidate = sample(10000L:60000L, 1L)
        result = tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server = result; port = candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn = socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      messages = list()
      new_page_count = 0L
      resize_sent = FALSE

      repeat {
        ready = socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line = tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg = tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages = c(messages, list(msg))

        # Respond to metrics_request
        if (identical(msg$type, "metrics_request")) {
          resp = if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(msg$str %||% "") * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        # Count newPage frames
        if (identical(msg$type, "frame") && isTRUE(msg$newPage)) {
          new_page_count = new_page_count + 1L
        }

        # After 2 newPage frames, send a normal resize (no plotIndex)
        if (new_page_count >= 2L && !resize_sent) {
          resize_sent = TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize", width = 500L, height = 400L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port = NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str = readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port = as.integer(port_str); break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status = bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("resize replay plotNumber matches the original plot's plotNumber", {
  skip_on_cran()

  server = start_mock_server_resize_plotnumber()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)

  # Plot 1
  plot(1:5)
  # Plot 2
  plot(1:10)

  # Wait for the mock server to receive the second newPage frame and send
  # back the resize message over the network.
  Sys.sleep(0.5)

  # Process the resize that the mock server sent after seeing 2 newPage frames.
  # The resize may have already been consumed by check_incoming during drawing,
  # so poll_resize returning FALSE is acceptable — what matters is whether
  # a resize replay frame was sent (checked below).
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs = server$collect()

  # Extract frames
  frames = Filter(function(m) identical(m$type, "frame"), msgs)

  # Find newPage frames — these are the original plot creations
  new_page_frames = Filter(function(f) isTRUE(f$newPage), frames)
  expect_true(length(new_page_frames) >= 2,
    info = "Should have at least 2 newPage frames")

  # The second newPage frame's plotNumber is the original plot 2's ID
  plot2_plotnumber = new_page_frames[[2]]$plotNumber
  expect_true(!is.null(plot2_plotnumber),
    info = "Plot 2's newPage frame should have a plotNumber")

  # Find resize replay frames
  resize_frames = Filter(function(f) isTRUE(f$resizeReplay), frames)
  expect_true(length(resize_frames) >= 1,
    info = "Should have at least 1 resize replay frame")

  resize_plotnumber = resize_frames[[1]]$plotNumber
  expect_true(!is.null(resize_plotnumber),
    info = "Resize replay frame should have a plotNumber")

  # THE KEY ASSERTION: resize replay's plotNumber must match the original
  # plot 2's plotNumber so the browser-side replaceLatest guard accepts it.
  expect_equal(resize_plotnumber, plot2_plotnumber,
    info = paste0(
      "Resize replay plotNumber (", resize_plotnumber,
      ") must match plot 2's plotNumber (", plot2_plotnumber,
      ") — if they differ, the browser rejects the resize replay"))
})

# Tests that a plotIndex resize replay frame contains the CORRECT historical
# plot's content, not the current (latest) plot's content.
#
# Bug: When viewing a historical plot and resizing, the replay frame should
# contain the historical plot's drawing ops.  With grid-based packages
# (ggplot2, ComplexHeatmap), the snapshot replay via GEplaySnapshot +
# grid.refresh() may produce the wrong plot's content if the grid display
# list is not correctly restored from the snapshot.
#
# This test creates two ggplot2 plots with distinct titles ("PLOT_AAA" and
# "PLOT_ZZZ"), sends a plotIndex=0 resize, and verifies the replay frame
# contains text ops with "PLOT_AAA" (plot 1's title), not "PLOT_ZZZ".

# TCP mock server that sends a plotIndex=0 resize after receiving two
# newPage frames, then collects all messages including the replay.
start_mock_server_plotindex_content = function() {
  skip_if_not_installed("callr")
  skip_if_not_installed("jsonlite")

  port_file = tempfile(pattern = "jgd-pi-content-port-", fileext = ".txt")

  bg = callr::r_bg(
    function(port_file) {
      `%||%` = function(x, y) if (is.null(x)) y else x
      safe_write = function(conn, text) {
        tryCatch(
          { writeLines(text, conn); flush(conn) },
          error = function(e) invisible(NULL)
        )
      }

      server = NULL; port = NULL
      for (i in seq_len(20)) {
        candidate = sample(10000L:60000L, 1L)
        result = tryCatch(serverSocket(candidate), error = function(e) NULL)
        if (!is.null(result)) { server = result; port = candidate; break }
      }
      if (is.null(port)) stop("Could not find free port")
      on.exit(close(server), add = TRUE)
      writeLines(as.character(port), port_file)

      conn = socketAccept(server, blocking = TRUE, open = "r+")
      on.exit(close(conn), add = TRUE)

      messages = list()
      new_page_count = 0L
      resize_sent = FALSE

      repeat {
        ready = socketSelect(list(conn), timeout = 5)
        if (!ready) next

        line = tryCatch(readLines(conn, n = 1), error = function(e) character(0))
        if (length(line) == 0 || !nzchar(line)) next

        msg = tryCatch(
          jsonlite::fromJSON(line, simplifyVector = FALSE),
          error = function(e) NULL
        )
        if (is.null(msg)) next
        messages = c(messages, list(msg))

        # Respond to metrics_request
        if (identical(msg$type, "metrics_request")) {
          resp = if (identical(msg$kind, "strWidth")) {
            list(type = "metrics_response", id = msg$id,
                 width = nchar(msg$str %||% "") * 8.0)
          } else {
            list(type = "metrics_response", id = msg$id,
                 ascent = 10.0, descent = 3.0, width = 8.0)
          }
          safe_write(conn, jsonlite::toJSON(resp, auto_unbox = TRUE))
        }

        # Count newPage frames
        if (identical(msg$type, "frame") && isTRUE(msg$newPage)) {
          new_page_count = new_page_count + 1L
        }

        # After 2 newPage frames, send a plotIndex=0 resize
        # (simulates browser viewing historical plot 1 and resizing)
        if (new_page_count >= 2L && !resize_sent) {
          resize_sent = TRUE
          safe_write(conn, jsonlite::toJSON(list(
            type = "resize", width = 500L, height = 400L, plotIndex = 0L
          ), auto_unbox = TRUE))
        }

        if (identical(msg$type, "close")) break
      }

      messages
    },
    args = list(port_file = port_file),
    supervise = TRUE
  )

  port = NULL
  for (i in seq_len(50)) {
    if (file.exists(port_file)) {
      port_str = readLines(port_file, n = 1, warn = FALSE)
      if (length(port_str) > 0 && nzchar(port_str)) {
        port = as.integer(port_str); break
      }
    }
    Sys.sleep(0.1)
  }
  if (is.null(port)) { bg$kill(); skip("Mock server did not start in time") }

  list(
    bg = bg,
    socket_url = sprintf("tcp://127.0.0.1:%d", port),
    collect = function(timeout = 15000) {
      bg$wait(timeout)
      status = bg$get_exit_status()
      if (!is.null(status) && status != 0) {
        stop("Mock server exited with error: ", bg$read_error())
      }
      if (is.null(status)) {
        bg$kill()
        skip("Mock server did not exit in time")
      }
      bg$get_result()
    },
    cleanup = function() {
      if (bg$is_alive()) bg$kill()
      unlink(port_file)
    }
  )
}

test_that("plotIndex resize replay contains the historical plot's content, not the current plot's", {
  skip_on_cran()
  skip_if_not_installed("ggplot2")

  server = start_mock_server_plotindex_content()
  withr::defer(server$cleanup())

  jgd(width = 4, height = 3, dpi = 72, socket = server$socket_url)

  # Plot 1 with a distinctive title
  set.seed(42)
  d1 = data.frame(x = runif(5), y = runif(5))
  print(ggplot2::ggplot(d1, ggplot2::aes(x, y)) +
    ggplot2::geom_point() +
    ggplot2::ggtitle("PLOT_AAA"))

  # Plot 2 with a different distinctive title
  set.seed(99)
  d2 = data.frame(x = runif(5), y = runif(5))
  print(ggplot2::ggplot(d2, ggplot2::aes(x, y)) +
    ggplot2::geom_point() +
    ggplot2::ggtitle("PLOT_ZZZ"))

  # Wait for the mock server to send the plotIndex=0 resize
  Sys.sleep(0.5)

  # Process the plotIndex=0 resize
  .Call(jgd:::C_jgd_poll_resize)

  dev.off()
  msgs = server$collect()

  # Find resize replay frames with plotIndex=0
  frames = Filter(function(m) identical(m$type, "frame"), msgs)
  resize_frames = Filter(
    function(f) isTRUE(f$resizeReplay) && identical(f$plotIndex, 0L),
    frames
  )
  expect_true(length(resize_frames) >= 1,
    info = "Should have at least 1 plotIndex=0 resize replay frame")

  # Extract all text ops from the replay frame
  replay_ops = resize_frames[[1]]$plot$ops
  text_ops = Filter(function(o) identical(o$op, "text"), replay_ops)
  text_strings = vapply(text_ops, function(o) if (is.null(o$str)) "" else o$str, character(1))

  # THE KEY ASSERTION: the replay frame must contain plot 1's title
  expect_true("PLOT_AAA" %in% text_strings,
    info = paste0(
      "plotIndex=0 resize replay should contain plot 1's title 'PLOT_AAA'. ",
      "Text ops found: [", paste(text_strings, collapse = ", "), "]"))

  # And must NOT contain plot 2's title
  expect_false("PLOT_ZZZ" %in% text_strings,
    info = paste0(
      "plotIndex=0 resize replay should NOT contain plot 2's title 'PLOT_ZZZ'. ",
      "Text ops found: [", paste(text_strings, collapse = ", "), "]"))
})

test_that("dev.off() sends close message", {
  msgs = with_mock_jgd({
    plot.new()
  })

  close_msgs = Filter(function(m) identical(m$type, "close"), msgs)
  expect_length(close_msgs, 1)
})

test_that("drawing after plot.new() generates a frame", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames = extract_frames(msgs)
  expect_true(length(frames) >= 1)
})

test_that("frame contains correct device dimensions", {
  msgs = with_mock_jgd(width = 4, height = 3, dpi = 72, {
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames = extract_frames(msgs)
  frame = frames[[1]]

  expect_equal(frame$plot$version, 1)
  expect_equal(frame$plot$device$width, 4 * 72)
  expect_equal(frame$plot$device$height, 3 * 72)
  expect_equal(frame$plot$device$dpi, 72)
})

test_that("rect() produces rect op", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0.2, 0.2, 0.8, 0.8)
  })

  rect_ops = extract_ops_by_type(msgs, "rect")
  expect_true(length(rect_ops) >= 1)
})

test_that("text() produces text op with correct string", {
  msgs = with_mock_jgd({
    plot.new()
    text(0.5, 0.5, "Hello jgd")
  })

  text_ops = extract_ops_by_type(msgs, "text")
  expect_true(length(text_ops) >= 1)
  expect_true(any(vapply(
    text_ops,
    function(o) identical(o$str, "Hello jgd"),
    logical(1)
  )))
})

test_that("circle() produces circle op", {
  msgs = with_mock_jgd({
    plot.new()
    symbols(0.5, 0.5, circles = 0.1, add = TRUE, inches = FALSE)
  })

  circle_ops = extract_ops_by_type(msgs, "circle")
  expect_true(length(circle_ops) >= 1)
})

test_that("lines() produces polyline op", {
  msgs = with_mock_jgd({
    plot.new()
    lines(c(0.2, 0.5, 0.8), c(0.3, 0.7, 0.3))
  })

  polyline_ops = extract_ops_by_type(msgs, "polyline")
  expect_true(length(polyline_ops) >= 1)
})

test_that("polygon() produces polygon op", {
  msgs = with_mock_jgd({
    plot.new()
    polygon(c(0.2, 0.5, 0.8), c(0.3, 0.8, 0.3))
  })

  polygon_ops = extract_ops_by_type(msgs, "polygon")
  expect_true(length(polygon_ops) >= 1)
})

test_that("gc fields are present in ops", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1, col = "red", border = "blue")
  })

  rect_ops = extract_ops_by_type(msgs, "rect")
  expect_true(length(rect_ops) >= 1)

  gc = rect_ops[[1]]$gc
  expect_true(!is.null(gc))
  expect_true(!is.null(gc$col))
  expect_true(!is.null(gc$fill))
  expect_true(!is.null(gc$lwd))
  expect_true(!is.null(gc$font))
})

test_that("clip op is generated", {
  msgs = with_mock_jgd({
    plot(1:3, 1:3)
  })

  clip_ops = extract_ops_by_type(msgs, "clip")
  expect_true(length(clip_ops) >= 1)
})

test_that("plot() generates multiple op types", {
  msgs = with_mock_jgd({
    plot(1:5, 1:5, main = "Test Plot")
  })

  ops = extract_ops(msgs)
  op_types = unique(vapply(ops, function(o) o$op, character(1)))

  # A basic scatter plot should generate at least these op types
  expect_true("clip" %in% op_types)
  expect_true("line" %in% op_types)
  expect_true("text" %in% op_types)
})

test_that("multiple pages generate multiple full frames", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1)
    plot.new()
    rect(0, 0, 0.5, 0.5)
  })

  frames = extract_frames(msgs)
  # At least 2 frames from the 2 pages (may have incremental frames too)
  expect_true(length(frames) >= 2)
})

test_that("metrics_request messages are sent for text", {
  msgs = with_mock_jgd({
    plot.new()
    text(0.5, 0.5, "metrics test")
  })

  metrics_msgs = Filter(
    function(m) identical(m$type, "metrics_request"),
    msgs
  )
  # text() triggers metrics requests (strWidth and/or metricInfo)
  expect_true(length(metrics_msgs) >= 1)
})

# --- Delta encoding tests ---

test_that("first flush on a page sends complete frame (incremental=false)", {
  msgs = with_mock_jgd({
    # plot.new triggers cb_newPage; drawing without hold triggers cb_mode flush
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames = extract_frames(msgs)
  expect_true(length(frames) >= 1)
  # The first frame must be a complete (non-incremental) frame
  expect_false(frames[[1]]$incremental)
})

test_that("subsequent flushes send only delta ops (incremental=true)", {
  msgs = with_mock_jgd({
    # First: a complete plot (plot.new + rect generates clip, rect, etc.)
    plot.new()
    rect(0, 0, 1, 1)
    # Additional drawing on the same page triggers incremental flush
    lines(c(0.2, 0.5, 0.8), c(0.2, 0.5, 0.8))
  })

  frames = extract_frames(msgs)
  # Should have at least 2 frames: one complete, one incremental
  expect_true(length(frames) >= 2)

  incr_frames = Filter(function(f) isTRUE(f$incremental), frames)
  expect_true(length(incr_frames) >= 1)

  # Delta encoding: incremental frame should contain only the new ops
  # (polyline from lines()), not ops from the first flush (rect, clip, etc.)
  incr_op_types = vapply(
    incr_frames[[1]]$plot$ops,
    function(o) o$op, character(1)
  )
  expect_true("polyline" %in% incr_op_types)
  expect_false("rect" %in% incr_op_types)
})

test_that("new page resets delta tracking", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1)
    lines(c(0.2, 0.5, 0.8), c(0.2, 0.5, 0.8))
    # Second page should reset delta tracking
    plot.new()
    rect(0, 0, 0.5, 0.5)
  })

  frames = extract_frames(msgs)
  # Find full (non-incremental) frames — should be at least 2 (one per page)
  full_frames = Filter(function(f) !isTRUE(f$incremental), frames)
  expect_true(length(full_frames) >= 2)
})

# --- Snapshot tests for JSON structure ---

test_that("close message JSON matches snapshot", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1)
  })

  close_msg = Filter(function(m) identical(m$type, "close"), msgs)[[1]]
  expect_json_snapshot(close_msg)
})

test_that("frame device metadata matches snapshot", {
  msgs = with_mock_jgd(width = 4, height = 3, dpi = 72, {
    plot.new()
    rect(0, 0, 1, 1)
  })

  frame = extract_frames(msgs)[[1]]
  # sessionId contains a process ID that varies — check format separately
  expect_match(frame$plot$sessionId, "^r-[0-9]+-[0-9]+$")
  expect_json_snapshot(list(
    version = frame$plot$version,
    device = frame$plot$device
  ))
})

test_that("rect op gc fields match snapshot", {
  msgs = with_mock_jgd({
    plot.new()
    rect(0, 0, 1, 1, col = "red", border = "blue", lwd = 2)
  })

  rect_ops = extract_ops_by_type(msgs, "rect")
  expect_json_snapshot(rect_ops[[1]]$gc)
})

test_that("rect coordinates use full double precision", {
  # Verify that coordinates are serialized with full double precision
  # (cJSON_CreateNumber) rather than truncated to 4 decimal places (%.4f).
  # Using 1/7 fractions produces repeating decimals that expose truncation.
  msgs = with_mock_jgd({
    par(mar = c(0, 0, 0, 0))
    plot.new()
    rect(1/3, 1/7, 2/3, 6/7)
  })

  rect_ops = extract_ops_by_type(msgs, "rect")
  op = rect_ops[[1]]
  # Use I(15) for 15 significant digits — enough to show precision beyond 4
  # decimal places while matching cJSON's %1.15g output format
  expect_snapshot(cat(jsonlite::toJSON(
    list(x0 = op$x0, y0 = op$y0, x1 = op$x1, y1 = op$y1),
    auto_unbox = TRUE, pretty = TRUE, digits = I(15)
  )))
})

test_that("text op structure matches snapshot", {
  msgs = with_mock_jgd({
    par(mar = c(0, 0, 0, 0))
    plot.new()
    text(0.5, 0.5, "snapshot test")
  })

  text_ops = extract_ops_by_type(msgs, "text")
  # Snapshot only stable fields (exclude coordinates which depend on layout)
  op = text_ops[[length(text_ops)]]
  expect_json_snapshot(list(
    op = op$op,
    str = op$str,
    rot = op$rot,
    hadj = op$hadj,
    gc = op$gc
  ))
})

test_that("TCP: basic connection and close", {
  msgs = with_mock_jgd(transport = "tcp", {
    plot.new()
    rect(0, 0, 1, 1)
  })

  close_msgs = Filter(function(m) identical(m$type, "close"), msgs)
  expect_length(close_msgs, 1)
})

test_that("TCP: frame generation with correct dimensions", {
  msgs = with_mock_jgd(width = 5, height = 4, dpi = 96, transport = "tcp", {
    plot.new()
    rect(0, 0, 1, 1)
  })

  frames = extract_frames(msgs)
  expect_true(length(frames) >= 1)

  frame = frames[[1]]
  expect_equal(frame$plot$device$width, 5 * 96)
  expect_equal(frame$plot$device$height, 4 * 96)
  expect_equal(frame$plot$device$dpi, 96)
})

test_that("TCP: metrics round-trip for text", {
  msgs = with_mock_jgd(transport = "tcp", {
    plot.new()
    text(0.5, 0.5, "tcp test")
  })

  metrics_msgs = Filter(
    function(m) identical(m$type, "metrics_request"),
    msgs
  )
  expect_true(length(metrics_msgs) >= 1)
})

test_that("TCP: multiple op types", {
  msgs = with_mock_jgd(transport = "tcp", {
    plot.new()
    rect(0.2, 0.2, 0.8, 0.8, col = "red")
    text(0.5, 0.5, "hello")
    lines(c(0.2, 0.8), c(0.3, 0.7))
  })

  ops = extract_ops(msgs)
  op_types = unique(vapply(ops, function(o) o$op, character(1)))
  expect_true("rect" %in% op_types)
  expect_true("text" %in% op_types)
  expect_true("polyline" %in% op_types)
})

test_that("TCP: gc fields match unix transport", {
  msgs = with_mock_jgd(transport = "tcp", {
    plot.new()
    rect(0, 0, 1, 1, col = "red", border = "blue", lwd = 2)
  })

  rect_ops = extract_ops_by_type(msgs, "rect")
  gc = rect_ops[[1]]$gc
  expect_true(!is.null(gc))
  expect_true(!is.null(gc$col))
  expect_true(!is.null(gc$fill))
  expect_true(!is.null(gc$lwd))
  expect_true(!is.null(gc$font))
})

#' jgd JSONL Protocol Specification
#'
#' @description
#' The jgd device communicates with a rendering server over JSONL
#' (JSON Lines). Messages are exchanged over a persistent
#' connection using one of three transport protocols: Unix domain sockets
#' (Linux/macOS), Windows named pipes, or TCP.
#'
#' This document specifies the wire protocol so that third-party servers
#' can implement a compatible rendering backend.
#'
#' @section Transport protocols:
#'
#' The client connects to the server using one of the following URI
#' schemes:
#'
#' - `unix:///path/to/socket` -- Unix domain socket (Linux/macOS
#'   default)
#' - `npipe:////./pipe/name` -- Windows named pipe (Windows default,
#'   Docker-standard 4-slash form)
#' - `tcp://host:port` -- TCP socket (any platform)
#'
#' Raw Unix socket paths (without a URI scheme) are also accepted.
#'
#' @section Message format:
#'
#' All messages are single-line JSON objects terminated by `\n`
#' (JSONL). Each message contains a `"type"` field identifying the
#' message kind. Encoding is always UTF-8.
#'
#' Receivers should ignore unknown top-level fields in any message
#' (forward-compatible). Unknown `"type"` values should be silently
#' discarded rather than treated as errors.
#'
#' @section Coordinate system:
#'
#' All coordinates in drawing operations are in **device pixels**
#' (i.e., `inches * dpi`). The origin `(0, 0)` is the **top-left**
#' corner of the device surface. The X axis increases to the right
#' and the Y axis increases downward.
#'
#' @section Connection handshake:
#'
#' The welcome message is **deferred**: the server waits until it
#' receives the first message from R before sending it. This avoids
#' a race condition on Windows named pipes where writing before the
#' first read completes can cause data loss.
#'
#' ```
#' R -> Server:  {"type":"ping"}
#' Server -> R:  {"type":"server_info", ...}
#' ```
#'
#' The server should also tolerate receiving a `frame` message
#' before `ping` (e.g., if a future client skips the ping). The
#' first received message of any type should trigger the deferred
#' welcome.
#'
#' @section Discovery file:
#'
#' The discovery file is an **optional** JSON file that allows the
#' client to find the server without an explicit socket address. It
#' is a hint for auto-connection only; the welcome message is the
#' single source of truth.
#'
#' **Location** (platform-specific):
#'
#' - Linux: `$XDG_CACHE_HOME/jgd/discovery.json`
#'   or `~/.cache/jgd/discovery.json`
#' - macOS: `~/Library/Caches/jgd/discovery.json`
#' - Windows: `%LOCALAPPDATA%/jgd/discovery.json`
#'
#' **Schema:**
#'
#' ```json
#' {
#'   "serverName": "jgd-http-server",
#'   "socketPath": "tcp://127.0.0.1:9000",
#'   "pid": 12345,
#'   "serverInfo": {
#'     "httpUrl": "http://127.0.0.1:8080/"
#'   }
#' }
#' ```
#'
#' - **`serverName`** (string, required): Human-readable server
#'   name.
#' - **`socketPath`** (string, required): Socket URI where the
#'   client should connect.
#' - **`pid`** (integer, required): Process ID of the server.
#' - **`serverInfo`** (object, optional): Flat key-value pairs with
#'   string values. Canonical key: `httpUrl` (HTTP endpoint URL).
#'
#' **Lifecycle:**
#'
#' - Written atomically (temp file + rename) after all listeners
#'   are ready.
#' - Server should remove the file on graceful shutdown, but only
#'   after confirming it still owns the file (PID check).
#' - Clients should verify liveness (e.g., PID check) before using
#'   stale files, since `~/.cache` is not cleared on reboot.
#' - Multiple server instances may coexist; the last writer wins.
#' - Server implementors may omit discovery file support entirely.
#'   Clients can always connect directly via an explicit socket
#'   URI.
#' - The discovery file has no explicit version field. Readers
#'   should ignore unknown fields for forward compatibility.
#'
#' @section server_info message:
#'
#' ```json
#' {
#'   "type": "server_info",
#'   "serverName": "jgd-http-server",
#'   "protocolVersion": 1,
#'   "transport": "unix",
#'   "serverInfo": {
#'     "httpUrl": "http://127.0.0.1:8080/"
#'   }
#' }
#' ```
#'
#' - **`type`**: `"server_info"` (string, always present)
#' - **`serverName`**: Human-readable server name (string)
#' - **`protocolVersion`**: Protocol version number, currently `1`
#'   (integer). Receivers should ignore messages with an unknown
#'   protocol version rather than raising an error.
#' - **`transport`**: Transport in use: `"tcp"`, `"unix"`, or
#'   `"npipe"` (string)
#' - **`serverInfo`**: A flat JSON object whose values are all
#'   strings (optional). Canonical key: `httpUrl`.
#'
#' See [jgd_server_info()] for how the R client represents this
#' data.
#'
#' @section R-to-server messages:
#'
#' **ping** -- Heartbeat; triggers the deferred welcome on first
#' send.
#'
#' ```json
#' {"type": "ping"}
#' ```
#'
#' **frame** -- A complete or incremental set of drawing operations.
#' See the Frame message section for the full schema.
#'
#' **metrics_request** -- Requests font metrics from the renderer.
#'
#' ```json
#' {"type": "metrics_request", "id": 1, "kind": "strWidth",
#'  "str": "Hello",
#'  "gc": {"font": {"family": "sans", "face": 1,
#'                   "size": 12}}}
#' ```
#'
#' ```json
#' {"type": "metrics_request", "id": 2, "kind": "metricInfo",
#'  "c": 77,
#'  "gc": {"font": {"family": "sans", "face": 1,
#'                   "size": 12}}}
#' ```
#'
#' - **`id`**: Request identifier (integer); the response must echo
#'   it.
#' - **`kind`**: `"strWidth"` (string width) or `"metricInfo"`
#'   (glyph metrics).
#' - **`str`** (string, `strWidth` only): The string to measure.
#' - **`c`** (integer, `metricInfo` only): Unicode code point of
#'   the character to measure (e.g., 77 for `"M"`).
#' - **`gc`**: Graphics context with a `font` object containing
#'   `family` (string), `face` (integer), and `size` (font size
#'   in points).
#'
#' **close** -- Signals device shutdown.
#'
#' ```json
#' {"type": "close"}
#' ```
#'
#' @section Server-to-R messages:
#'
#' **server_info** -- Welcome message (see above).
#'
#' **resize** -- Renderer viewport change.
#'
#' ```json
#' {"type": "resize", "width": 800, "height": 600}
#' ```
#'
#' - **`width`**, **`height`**: New viewport dimensions in device
#'   pixels (positive integers). These become the device's width
#'   and height directly (no DPI scaling is applied).
#' - **`plotIndex`** (integer, optional): If present, replay the
#'   historical plot identified by its R-assigned plot number (the
#'   `plotNumber` from earlier frames) instead of the current plot.
#'
#' **metrics_response** -- Font metrics from the renderer.
#'
#' ```json
#' {"type": "metrics_response", "id": 1, "width": 48.5,
#'  "ascent": 10.2, "descent": 2.8}
#' ```
#'
#' - **`id`**: Must match the request `id`.
#' - Servers should respond promptly. Clients handle their own
#'   timeouts and may fall back to local computation. Servers are
#'   not required to synthesize fallback responses.
#'
#' @section Frame message:
#'
#' The frame message carries drawing operations from R to the
#' server.
#'
#' New plot example:
#'
#' ```json
#' {
#'   "type": "frame",
#'   "incremental": false,
#'   "newPage": true,
#'   "plotNumber": 0,
#'   "ext": {},
#'   "plot": {
#'     "version": 1,
#'     "sessionId": "r-1234-1",
#'     "device": {
#'       "width": 768,
#'       "height": 576,
#'       "dpi": 96,
#'       "bg": "rgba(255,255,255,1)"
#'     },
#'     "ops": []
#'   }
#' }
#' ```
#'
#' (Minimal example; real frames typically start with a `clip` op.)
#'
#' Historical resize replay example:
#'
#' ```json
#' {
#'   "type": "frame",
#'   "incremental": false,
#'   "resizeReplay": true,
#'   "plotIndex": 0,
#'   "plot": { "..." }
#' }
#' ```
#'
#' **Top-level fields:**
#'
#' - **`type`**: `"frame"` (always present).
#' - **`incremental`** (boolean, always present): If `true`, `ops`
#'   contains only operations added since the last flush (delta).
#'   If `false`, `ops` contains the complete drawing for the page.
#' - **`newPage`** (boolean, optional): Present and `true` when
#'   this is a fresh plot (not a delta, not a resize replay).
#' - **`resizeReplay`** (boolean, optional): Present and `true`
#'   when this frame is a replay triggered by a resize.
#' - **`plotIndex`** (integer, optional): Present during
#'   `resizeReplay` when a historical plot (not the current one)
#'   was replayed. This is the absolute R-side plot number (the
#'   same 0-based value previously sent as `plotNumber` when the
#'   plot was created). It may diverge from the renderer's current
#'   history array index after deletions or evictions. See also
#'   the Resize protocol section for the full resize flow.
#' - **`plotNumber`** (integer, optional): Absolute 0-based
#'   sequence number for plots (e.g., 0 for the first, 1 for the
#'   second). Present on all frames for the current plot (including
#'   incremental and resize replay frames). Omitted only on
#'   historical resize replays where `plotIndex` is present.
#' - **`ext`** (object, optional): Frame-level extension data.
#'   When unset, the field is omitted (never sent as `null`); when
#'   set, it may be any JSON object including an empty `{}`.
#'   Servers should preserve and forward it to renderers.
#'
#' **plot object:**
#'
#' - **`version`** (integer): Protocol version, currently `1`.
#' - **`sessionId`** (string): Identifies the R session/device.
#'   Used for routing resize requests to the correct R process.
#' - **`device`** (object):
#'   - **`width`**, **`height`**: Device dimensions in pixels.
#'   - **`dpi`**: Dots per inch.
#'   - **`bg`**: Background color as an RGBA string
#'     (see the Color format section).
#' - **`ops`** (array): Drawing operations
#'   (see the Drawing operations section).
#'
#' @section Color format:
#'
#' Colors are represented as CSS-style RGBA strings:
#'
#' ```
#' "rgba(R,G,B,A)"
#' ```
#'
#' - R, G, B: integers 0--255.
#' - A: decimal 0.0--1.0 (e.g., `"rgba(0,0,0,0.502)"`).
#'
#' Transparent or `NA` colors are represented as JSON `null`.
#'
#' @section Graphics context:
#'
#' Most drawing operations include a `"gc"` object (exceptions are
#' noted per operation):
#'
#' ```json
#' {
#'   "col": "rgba(0,0,0,1)",
#'   "fill": null,
#'   "lwd": 1.0,
#'   "lty": [],
#'   "lend": "round",
#'   "ljoin": "round",
#'   "lmitre": 10.0,
#'   "font": {
#'     "family": "sans",
#'     "face": 1,
#'     "size": 12.0,
#'     "lineheight": 1.2
#'   },
#'   "ext": {}
#' }
#' ```
#'
#' - **`col`**: Stroke color (RGBA string or `null`).
#' - **`fill`**: Fill color (RGBA string or `null`).
#' - **`lwd`**: Line width in pixels (number).
#' - **`lty`**: Line type as an array of dash lengths (in pixels).
#'   Solid lines and blank (invisible) lines both produce an empty
#'   array `[]`. When the line is blank, `col` is `null`, so
#'   renderers can distinguish via the color.
#' - **`lend`**: Line end cap: `"round"`, `"butt"`, or `"square"`.
#' - **`ljoin`**: Line join: `"round"`, `"miter"`, or `"bevel"`.
#' - **`lmitre`**: Miter limit (number).
#' - **`font`**:
#'   - **`family`**: Font family name (string; empty string if
#'     default).
#'   - **`face`**: Font face: 1 = plain, 2 = bold, 3 = italic,
#'     4 = bold italic, 5 = symbol.
#'   - **`size`**: Font size in points (number).
#'   - **`lineheight`**: Line height multiplier (number).
#' - **`ext`** (object, optional): Per-operation extension data.
#'   Present only when set. Free-form JSON.
#'
#' @section Drawing operations:
#'
#' Each element of the `ops` array is a JSON object with an `"op"`
#' field. Most drawing operations include a `"gc"` field (see
#' the Graphics context section). Exceptions are noted per operation.
#'
#' All coordinates are in device pixels with a top-left origin (see
#' the Coordinate system section).
#'
#' **clip** -- Set the clipping rectangle. No `gc`.
#'
#' ```json
#' {"op": "clip", "x0": 0, "y0": 0, "x1": 768, "y1": 576}
#' ```
#'
#' **line** -- A single line segment.
#'
#' ```json
#' {"op": "line", "x1": 100, "y1": 200,
#'  "x2": 300, "y2": 400, "gc": {}}
#' ```
#'
#' **polyline** -- Connected line segments (not closed).
#'
#' ```json
#' {"op": "polyline", "x": [1, 2, 3],
#'  "y": [4, 5, 6], "gc": {}}
#' ```
#'
#' **polygon** -- Closed polygon (filled and/or stroked).
#'
#' ```json
#' {"op": "polygon", "x": [1, 2, 3],
#'  "y": [4, 5, 6], "gc": {}}
#' ```
#'
#' **rect** -- Rectangle.
#'
#' ```json
#' {"op": "rect", "x0": 10, "y0": 20,
#'  "x1": 100, "y1": 80, "gc": {}}
#' ```
#'
#' **circle** -- Circle.
#'
#' ```json
#' {"op": "circle", "x": 50, "y": 50, "r": 25, "gc": {}}
#' ```
#'
#' **text** -- Text string.
#'
#' ```json
#' {"op": "text", "x": 100, "y": 200, "str": "Hello",
#'  "rot": 0, "hadj": 0.5, "gc": {}}
#' ```
#'
#' - **`str`**: The text content (string).
#' - **`rot`**: Rotation angle in degrees (counter-clockwise).
#' - **`hadj`**: Horizontal adjustment (0 = left-aligned,
#'   0.5 = centered, 1 = right-aligned).
#'
#' **path** -- Complex path with subpaths and a fill rule.
#'
#' ```json
#' {"op": "path", "winding": "nonzero",
#'  "subpaths": [[[10, 20], [30, 40], [50, 20]]],
#'  "gc": {}}
#' ```
#'
#' - **`winding`**: Fill rule: `"nonzero"` or `"evenodd"`.
#' - **`subpaths`**: Array of subpaths. Each subpath is an array
#'   of `[x, y]` coordinate pairs.
#'
#' **raster** -- Raster image. No `gc`.
#'
#' ```json
#' {"op": "raster", "x": 0, "y": 576, "w": 100, "h": -80,
#'  "rot": 0, "interpolate": true,
#'  "pw": 200, "ph": 160,
#'  "data": "data:image/png;base64,..."}
#' ```
#'
#' - **`x`**, **`y`**: Bottom-left corner of the destination
#'   rectangle in device coordinates.
#' - **`w`**, **`h`**: Displayed width and height. May be
#'   **negative** to indicate a horizontal or vertical flip;
#'   renderers should use the absolute value for sizing and adjust
#'   the anchor point accordingly.
#' - **`pw`**, **`ph`**: Pixel width and height of the source
#'   image.
#' - **`rot`**: Rotation angle in degrees.
#' - **`interpolate`**: Whether to interpolate when scaling.
#' - **`data`**: Base64-encoded PNG as a data URI.
#'
#' **beginGroup** -- Start a drawing group (experimental). No `gc`.
#'
#' ```json
#' {"op": "beginGroup",
#'  "ext": {"filter": "blur(5px)", "opacity": 0.8}}
#' ```
#'
#' - **`ext`** (object, optional): Group-level extension data.
#'   Present only when set. Free-form JSON. Common keys: `filter`
#'   (CSS filter string), `opacity` (number 0--1), `blendMode`
#'   (CSS blend mode string).
#'
#' **endGroup** -- End the most recently opened group. No `gc`, no
#' fields other than `"op"`.
#'
#' ```json
#' {"op": "endGroup"}
#' ```
#'
#' Groups nest arbitrarily.
#'
#' @section Resize protocol:
#'
#' The server receives resize messages from the renderer and
#' forwards them to R. R replays the drawing at the new dimensions
#' and sends back a frame message.
#'
#' **Normal resize flow:**
#'
#' ```
#' Renderer -> Server:  {"type":"resize","width":800,
#'                       "height":600}
#' Server   -> R:       {"type":"resize","width":800,
#'                       "height":600}
#' R        -> Server:  {"type":"frame",
#'                       "resizeReplay":true,
#'                       "incremental":false,...}
#' ```
#'
#' **History resize flow** (replay a historical plot):
#'
#' ```
#' Renderer -> Server:  {"type":"resize","width":800,
#'                       "height":600,"plotIndex":2,
#'                       "sessionId":"r-1234-1"}
#' Server   -> R:       {"type":"resize","width":800,
#'                       "height":600,"plotIndex":2}
#' R        -> Server:  {"type":"frame",
#'                       "resizeReplay":true,
#'                       "plotIndex":2,
#'                       "incremental":false,...}
#' ```
#'
#' Note: The server strips `sessionId` before forwarding to R.
#' History resizes are routed only to the R session that owns the
#' target plot.
#'
#' **Resize deduplication:**
#'
#' Servers should deduplicate consecutive normal resizes with
#' identical dimensions for each R session. However, if the
#' previous resize was a `plotIndex` resize, the next normal resize
#' at the same dimensions must NOT be deduplicated, because they
#' target different contexts (historical snapshot vs. current
#' plot).
#'
#' @section Multiple R sessions:
#'
#' A server may accept connections from multiple R processes
#' simultaneously. Each R connection has its own `sessionId` and
#' independent state. Servers should:
#'
#' - Route `metrics_request` messages from each R connection to
#'   the renderer, and route the matching `metrics_response` back
#'   to the originating R session. The `id` field is only unique
#'   within a single R process, so servers must scope routing by
#'   the originating connection (e.g. keyed by session + `id`).
#'   If a server remaps IDs when forwarding to a shared renderer,
#'   it must restore the original `id` when relaying the response
#'   back to R.
#' - Route `plotIndex` resizes to the R session that owns the
#'   target plot (identified by `sessionId` in the resize message).
#' - Broadcast normal resizes to all connected R sessions.
#' - Broadcast `frame` and `close` messages to all connected
#'   renderers.
#'
#' @section Session ID management:
#'
#' The `sessionId` in frame messages identifies the R device
#' instance. Servers should treat it as an **opaque string**. Do
#' not parse it or make assumptions about its format.
#'
#' Session ID reuse is unlikely but possible (e.g., after process
#' restart). As a defensive measure, servers should disambiguate
#' if a previously retired `sessionId` reappears, to prevent
#' `plotIndex` resizes for old plots from reaching the new
#' connection. The server's (possibly remapped) `sessionId` is
#' what the renderer sees; `plotIndex` resizes use it for routing.
#'
#' @section Connection lifecycle:
#'
#' **Graceful close:** The client sends `{"type":"close"}` on
#' device shutdown. The server should forward this to renderers
#' and clean up routing state for that session.
#'
#' **Ungraceful disconnect:** If the connection drops without a
#' `close` message (e.g., process crash), the server should detect
#' the broken connection (EOF or socket error), clean up the
#' session, and optionally notify renderers.
#'
#' **Incomplete lines:** If a connection drops mid-line (no
#' trailing `\n`), the partial data should be discarded.
#'
#' @section Extension fields:
#'
#' Extension fields (`ext`) appear at three levels:
#'
#' - **Frame-level**: Top-level `ext` on the frame message.
#'   Applies to the entire frame.
#' - **Graphics context level**: `ext` inside the `gc` object.
#'   Applies to all drawing operations while active.
#' - **Group level**: `ext` on `beginGroup` operations. Applies
#'   only to that group.
#'
#' All `ext` fields are free-form JSON objects. When unset, the
#' field is omitted from the message (never sent as `null`). When
#' set, it may contain any JSON object, including an empty `{}`.
#' Servers should preserve and forward them to renderers without
#' validation. Renderers should ignore unknown keys.
#'
#' Extension fields are preserved across resize replays, so
#' historical plot snapshots retain their `ext` data.
#'
#' See [jgd_ext()], [jgd_frame_ext()], and [jgd_begin_group()]
#' for the R API to set these fields.
#'
#' @section Implementing a server:
#'
#' A minimal server implementation needs to:
#'
#' 1. Listen on a Unix socket, named pipe, or TCP port.
#' 2. Accept R connections and read JSONL lines.
#' 3. Send a deferred `server_info` welcome after receiving the
#'    first message from R (not before).
#' 4. Forward `frame` messages to connected renderers.
#' 5. Forward `resize` messages from renderers to R.
#' 6. Handle `metrics_request`/`metrics_response` routing between
#'    R and the renderer. Clients handle their own timeouts;
#'    servers are not required to synthesize fallback responses.
#' 7. Forward `close` messages to renderers and clean up the
#'    session state (remove metrics routing entries, etc.).
#'
#' Optional:
#'
#' - Write a discovery file on startup, remove on shutdown.
#' - Implement resize deduplication.
#' - Implement session ID remapping for reused IDs.
#' - Serve an HTTP endpoint for the renderer UI.
#'
#' @name jgd_spec
#' @aliases jgd_protocol
#' @seealso [jgd()], [jgd_server_info()], [jgd_ext()],
#'   [jgd_frame_ext()], [jgd_begin_group()]
NULL

# TODO: Message size limits are intentionally unspecified.  The
# reference R client uses a 4096-byte read buffer (transport.h),
# which may truncate very large messages (e.g., raster data URIs).
# This is an implementation limitation, not a protocol constraint.
# A future protocol version may address this if needed.

# Unix: jgd() opens a device and dev.off() closes it

    Code
      jgd(socket = "unix:///nonexistent-jgd-test.sock")
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# Unix: jgd() respects JGD_SOCKET env var

    Code
      jgd()
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# Unix: jgd() respects jgd.socket option

    Code
      jgd()
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# Unix: drawing works without server connection

    Code
      jgd(socket = "unix:///nonexistent-jgd-draw-test.sock")
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# Unix: unix://localhost/ URI accepted

    Code
      jgd(socket = "unix://localhost/nonexistent-jgd-localhost-test.sock")
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# Unix: drawing works with unix://localhost/ URI

    Code
      jgd(socket = "unix://localhost/nonexistent-jgd-localhost-draw.sock")
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# TCP: jgd() opens a device and dev.off() closes it

    Code
      jgd(socket = "tcp://127.0.0.1:1")
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# TCP: jgd() respects JGD_SOCKET env var

    Code
      jgd()
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# TCP: jgd() respects jgd.socket option

    Code
      jgd()
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# TCP: drawing works without server connection

    Code
      jgd(socket = "tcp://127.0.0.1:1")
    Condition
      Warning in `jgd()`:
      jgd: could not connect to renderer. Plots will be recorded but not displayed until connection is established.

# jgd() validates socket parameter type

    Code
      jgd(socket = 42)
    Condition
      Error in `jgd()`:
      ! is.character(socket) is not TRUE

---

    Code
      jgd(socket = c("a", "b"))
    Condition
      Error in `jgd()`:
      ! length(socket) == 1L is not TRUE

# malformed JSON: device metadata matches snapshot

    Code
      cat(json)
    Output
      {
        "version": 1,
        "device": {
          "width": 288,
          "height": 216,
          "dpi": 72,
          "bg": "rgba(255,255,255,1)"
        }
      }

# malformed JSON: rect gc fields match snapshot

    Code
      cat(json)
    Output
      {
        "col": "rgba(0,0,255,1)",
        "fill": "rgba(255,0,0,1)",
        "lwd": 2,
        "lty": [],
        "lend": "round",
        "ljoin": "round",
        "lmitre": 10,
        "font": {
          "family": "",
          "face": 1,
          "size": 12,
          "lineheight": 1
        }
      }

# close message JSON matches snapshot

    Code
      cat(json)
    Output
      {
        "type": "close"
      }

# frame device metadata matches snapshot

    Code
      cat(json)
    Output
      {
        "version": 1,
        "device": {
          "width": 288,
          "height": 216,
          "dpi": 72,
          "bg": "rgba(255,255,255,1)"
        }
      }

# rect op gc fields match snapshot

    Code
      cat(json)
    Output
      {
        "col": "rgba(0,0,255,1)",
        "fill": "rgba(255,0,0,1)",
        "lwd": 2,
        "lty": [],
        "lend": "round",
        "ljoin": "round",
        "lmitre": 10,
        "font": {
          "family": "",
          "face": 1,
          "size": 12,
          "lineheight": 1
        }
      }

# rect coordinates use full double precision

    Code
      cat(jsonlite::toJSON(list(x0 = op$x0, y0 = op$y0, x1 = op$x1, y1 = op$y1),
      auto_unbox = TRUE, pretty = TRUE, digits = I(15)))
    Output
      {
        "x0": 99.5555555555555,
        "y0": 179.428571428571,
        "x1": 188.444444444444,
        "y1": 36.5714285714286
      }

# text op structure matches snapshot

    Code
      cat(json)
    Output
      {
        "op": "text",
        "str": "snapshot test",
        "rot": 0,
        "hadj": 0.5,
        "gc": {
          "col": "rgba(0,0,0,1)",
          "fill": "rgba(255,255,255,1)",
          "lwd": 1,
          "lty": [],
          "lend": "round",
          "ljoin": "round",
          "lmitre": 10,
          "font": {
            "family": "",
            "face": 1,
            "size": 12,
            "lineheight": 1
          }
        }
      }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd_end_group}
\alias{jgd_end_group}
\title{End a drawing group (experimental)}
\usage{
jgd_end_group()
}
\value{
Called for its side effect; returns \code{NULL} invisibly.
}
\description{
Emits an \code{endGroup} operation into the drawing stream, closing the most
recently opened group from \code{\link[=jgd_begin_group]{jgd_begin_group()}}.
}
\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd_server_info}
\alias{jgd_server_info}
\title{Get server information}
\usage{
jgd_server_info()
}
\value{
A named list, or \code{NULL}.

When connected:
\itemize{
\item \strong{\code{connected}}: \code{TRUE}
\item \strong{\code{server_name}}: Server name (character)
\item \strong{\code{protocol_version}}: Protocol version (integer)
\item \strong{\code{transport}}: Transport protocol (character)
\item \strong{\code{server_info}}: Named character vector of key-value pairs
from the server's \code{serverInfo} object
(e.g. \code{c(httpUrl = "http://...")}); empty if absent
}

When not connected (discovery file fallback):
\itemize{
\item \strong{\code{connected}}: \code{FALSE}
\item \strong{\code{server_name}}: Server name (character)
\item \strong{\code{socket_path}}: Socket URI (character)
\item \strong{\code{pid}}: Server process ID (integer)
\item \strong{\code{server_info}}: Named character vector (as above)
}
}
\description{
Returns metadata about the jgd server. When a jgd device is open
and connected, returns the welcome message information with
\code{connected = TRUE}. Otherwise, falls back to reading the
discovery file and returns information with \code{connected = FALSE}.
Returns \code{NULL} if no information is available from either source.
}
\details{
The discovery fallback applies regardless of whether the current
device is a jgd device. This means \code{jgd_server_info()} can
return a non-\code{NULL} result even when no jgd device is open, as
long as a valid discovery file exists.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{with_jgd_ext}
\alias{with_jgd_ext}
\title{Scoped extended graphics context (experimental)}
\usage{
with_jgd_ext(json, expr)
}
\arguments{
\item{json}{A single JSON string representing the extension object.
Must be valid JSON; an error is raised otherwise.  Unlike \code{\link[=jgd_ext]{jgd_ext()}},
\code{NULL} is not accepted (use \code{jgd_ext(NULL)} to clear ext explicitly).}

\item{expr}{Expression to evaluate with the extension active.}
}
\value{
The result of evaluating \code{expr}.
}
\description{
Temporarily sets extension fields for the duration of \code{expr}, then clears
ext on exit (sets to \code{NULL}).
}
\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd}
\alias{jgd}
\title{JSON Graphics Device}
\usage{
jgd(width = 8, height = 6, dpi = 96, socket = NULL)
}
\arguments{
\item{width}{Device width in inches (default 8).}

\item{height}{Device height in inches (default 6).}

\item{dpi}{Resolution in dots per inch (default 96).}

\item{socket}{Socket address for the rendering server. Supports URI formats
(\verb{tcp://host:port}, \verb{unix:///path/to/socket}) or raw Unix socket paths.
If \code{NULL} (default), use the \code{jgd.socket} R option, falling back to the
\code{JGD_SOCKET}environment variable. If \code{JGD_SOCKET} environment variable is
also unset, the device discovers the socket via the discovery file.}
}
\value{
Invisible \code{NULL}. The device is opened as a side effect.
}
\description{
Opens a graphics device that streams plot operations as JSON to an external
renderer (e.g. VS Code extension or browser) over a Unix domain socket.
}
\section{Displaying plots with \code{jgd}}{

It is important to note that \code{jgd()} does not display any plots; it only
streams them (i.e., converts them to a format that a JSON renderer
understands). To actually \emph{display} your plots with \code{jgd}, you'll need an
appropriate frontend. We provide two official renderers, both available for
install from the project repository:
\url{https://github.com/grantmcdermott/jgd}.
\itemize{
\item \strong{VS Code extension.} An integrated plot pane for VS Code.
\item \strong{Deno server.} A standalone browser-based renderer.
}

Users aren't limited to these two options. The \code{jgd} protocol is deliberately
frontend-agnostic; you can render plots with any client that reads JSONL
(JSON Lines). Again, please see the project repository for full documentation:
\url{https://github.com/grantmcdermott/jgd}
}

\section{Debugging}{

Set \code{options(jgd.debug = TRUE)} before opening the device to enable
frame-level diagnostic output on stderr (via \code{REprintf}).  This logs
details about \code{newPage}, \code{flush_frame}, and \code{poll_resize} events, which
is useful for diagnosing resize/replay issues.
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
# Requires a running renderer (e.g., VS Code extension or Deno server).
# See the "Displaying plots" section above.
library(jgd)
jgd()
plot(1:10)
lines(1:10, col = "red", lwd = 3)
hist(rnorm(1000), col = "steelblue")
dev.off()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spec.R
\name{jgd_spec}
\alias{jgd_spec}
\alias{jgd_protocol}
\title{jgd JSONL Protocol Specification}
\description{
The jgd device communicates with a rendering server over JSONL
(JSON Lines). Messages are exchanged over a persistent
connection using one of three transport protocols: Unix domain sockets
(Linux/macOS), Windows named pipes, or TCP.

This document specifies the wire protocol so that third-party servers
can implement a compatible rendering backend.
}
\section{Transport protocols}{


The client connects to the server using one of the following URI
schemes:
\itemize{
\item \verb{unix:///path/to/socket} -- Unix domain socket (Linux/macOS
default)
\item \verb{npipe:////./pipe/name} -- Windows named pipe (Windows default,
Docker-standard 4-slash form)
\item \verb{tcp://host:port} -- TCP socket (any platform)
}

Raw Unix socket paths (without a URI scheme) are also accepted.
}

\section{Message format}{


All messages are single-line JSON objects terminated by \verb{\\n}
(JSONL). Each message contains a \code{"type"} field identifying the
message kind. Encoding is always UTF-8.

Receivers should ignore unknown top-level fields in any message
(forward-compatible). Unknown \code{"type"} values should be silently
discarded rather than treated as errors.
}

\section{Coordinate system}{


All coordinates in drawing operations are in \strong{device pixels}
(i.e., \code{inches * dpi}). The origin \verb{(0, 0)} is the \strong{top-left}
corner of the device surface. The X axis increases to the right
and the Y axis increases downward.
}

\section{Connection handshake}{


The welcome message is \strong{deferred}: the server waits until it
receives the first message from R before sending it. This avoids
a race condition on Windows named pipes where writing before the
first read completes can cause data loss.

\if{html}{\out{<div class="sourceCode">}}\preformatted{R -> Server:  \{"type":"ping"\}
Server -> R:  \{"type":"server_info", ...\}
}\if{html}{\out{</div>}}

The server should also tolerate receiving a \code{frame} message
before \code{ping} (e.g., if a future client skips the ping). The
first received message of any type should trigger the deferred
welcome.
}

\section{Discovery file}{


The discovery file is an \strong{optional} JSON file that allows the
client to find the server without an explicit socket address. It
is a hint for auto-connection only; the welcome message is the
single source of truth.

\strong{Location} (platform-specific):
\itemize{
\item Linux: \verb{$XDG_CACHE_HOME/jgd/discovery.json}
or \verb{~/.cache/jgd/discovery.json}
\item macOS: \verb{~/Library/Caches/jgd/discovery.json}
\item Windows: \verb{\%LOCALAPPDATA\%/jgd/discovery.json}
}

\strong{Schema:}

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{
  "serverName": "jgd-http-server",
  "socketPath": "tcp://127.0.0.1:9000",
  "pid": 12345,
  "serverInfo": \{
    "httpUrl": "http://127.0.0.1:8080/"
  \}
\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{serverName}} (string, required): Human-readable server
name.
\item \strong{\code{socketPath}} (string, required): Socket URI where the
client should connect.
\item \strong{\code{pid}} (integer, required): Process ID of the server.
\item \strong{\code{serverInfo}} (object, optional): Flat key-value pairs with
string values. Canonical key: \code{httpUrl} (HTTP endpoint URL).
}

\strong{Lifecycle:}
\itemize{
\item Written atomically (temp file + rename) after all listeners
are ready.
\item Server should remove the file on graceful shutdown, but only
after confirming it still owns the file (PID check).
\item Clients should verify liveness (e.g., PID check) before using
stale files, since \verb{~/.cache} is not cleared on reboot.
\item Multiple server instances may coexist; the last writer wins.
\item Server implementors may omit discovery file support entirely.
Clients can always connect directly via an explicit socket
URI.
\item The discovery file has no explicit version field. Readers
should ignore unknown fields for forward compatibility.
}
}

\section{server_info message}{


\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{
  "type": "server_info",
  "serverName": "jgd-http-server",
  "protocolVersion": 1,
  "transport": "unix",
  "serverInfo": \{
    "httpUrl": "http://127.0.0.1:8080/"
  \}
\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{type}}: \code{"server_info"} (string, always present)
\item \strong{\code{serverName}}: Human-readable server name (string)
\item \strong{\code{protocolVersion}}: Protocol version number, currently \code{1}
(integer). Receivers should ignore messages with an unknown
protocol version rather than raising an error.
\item \strong{\code{transport}}: Transport in use: \code{"tcp"}, \code{"unix"}, or
\code{"npipe"} (string)
\item \strong{\code{serverInfo}}: A flat JSON object whose values are all
strings (optional). Canonical key: \code{httpUrl}.
}

See \code{\link[=jgd_server_info]{jgd_server_info()}} for how the R client represents this
data.
}

\section{R-to-server messages}{


\strong{ping} -- Heartbeat; triggers the deferred welcome on first
send.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"type": "ping"\}
}\if{html}{\out{</div>}}

\strong{frame} -- A complete or incremental set of drawing operations.
See the Frame message section for the full schema.

\strong{metrics_request} -- Requests font metrics from the renderer.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"type": "metrics_request", "id": 1, "kind": "strWidth",
 "str": "Hello",
 "gc": \{"font": \{"family": "sans", "face": 1,
                  "size": 12\}\}\}
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"type": "metrics_request", "id": 2, "kind": "metricInfo",
 "c": 77,
 "gc": \{"font": \{"family": "sans", "face": 1,
                  "size": 12\}\}\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{id}}: Request identifier (integer); the response must echo
it.
\item \strong{\code{kind}}: \code{"strWidth"} (string width) or \code{"metricInfo"}
(glyph metrics).
\item \strong{\code{str}} (string, \code{strWidth} only): The string to measure.
\item \strong{\code{c}} (integer, \code{metricInfo} only): Unicode code point of
the character to measure (e.g., 77 for \code{"M"}).
\item \strong{\code{gc}}: Graphics context with a \code{font} object containing
\code{family} (string), \code{face} (integer), and \code{size} (font size
in points).
}

\strong{close} -- Signals device shutdown.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"type": "close"\}
}\if{html}{\out{</div>}}
}

\section{Server-to-R messages}{


\strong{server_info} -- Welcome message (see above).

\strong{resize} -- Renderer viewport change.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"type": "resize", "width": 800, "height": 600\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{width}}, \strong{\code{height}}: New viewport dimensions in device
pixels (positive integers). These become the device's width
and height directly (no DPI scaling is applied).
\item \strong{\code{plotIndex}} (integer, optional): If present, replay the
historical plot identified by its R-assigned plot number (the
\code{plotNumber} from earlier frames) instead of the current plot.
}

\strong{metrics_response} -- Font metrics from the renderer.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"type": "metrics_response", "id": 1, "width": 48.5,
 "ascent": 10.2, "descent": 2.8\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{id}}: Must match the request \code{id}.
\item Servers should respond promptly. Clients handle their own
timeouts and may fall back to local computation. Servers are
not required to synthesize fallback responses.
}
}

\section{Frame message}{


The frame message carries drawing operations from R to the
server.

New plot example:

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{
  "type": "frame",
  "incremental": false,
  "newPage": true,
  "plotNumber": 0,
  "ext": \{\},
  "plot": \{
    "version": 1,
    "sessionId": "r-1234-1",
    "device": \{
      "width": 768,
      "height": 576,
      "dpi": 96,
      "bg": "rgba(255,255,255,1)"
    \},
    "ops": []
  \}
\}
}\if{html}{\out{</div>}}

(Minimal example; real frames typically start with a \code{clip} op.)

Historical resize replay example:

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{
  "type": "frame",
  "incremental": false,
  "resizeReplay": true,
  "plotIndex": 0,
  "plot": \{ "..." \}
\}
}\if{html}{\out{</div>}}

\strong{Top-level fields:}
\itemize{
\item \strong{\code{type}}: \code{"frame"} (always present).
\item \strong{\code{incremental}} (boolean, always present): If \code{true}, \code{ops}
contains only operations added since the last flush (delta).
If \code{false}, \code{ops} contains the complete drawing for the page.
\item \strong{\code{newPage}} (boolean, optional): Present and \code{true} when
this is a fresh plot (not a delta, not a resize replay).
\item \strong{\code{resizeReplay}} (boolean, optional): Present and \code{true}
when this frame is a replay triggered by a resize.
\item \strong{\code{plotIndex}} (integer, optional): Present during
\code{resizeReplay} when a historical plot (not the current one)
was replayed. This is the absolute R-side plot number (the
same 0-based value previously sent as \code{plotNumber} when the
plot was created). It may diverge from the renderer's current
history array index after deletions or evictions. See also
the Resize protocol section for the full resize flow.
\item \strong{\code{plotNumber}} (integer, optional): Absolute 0-based
sequence number for plots (e.g., 0 for the first, 1 for the
second). Present on all frames for the current plot (including
incremental and resize replay frames). Omitted only on
historical resize replays where \code{plotIndex} is present.
\item \strong{\code{ext}} (object, optional): Frame-level extension data.
When unset, the field is omitted (never sent as \code{null}); when
set, it may be any JSON object including an empty \code{{}}.
Servers should preserve and forward it to renderers.
}

\strong{plot object:}
\itemize{
\item \strong{\code{version}} (integer): Protocol version, currently \code{1}.
\item \strong{\code{sessionId}} (string): Identifies the R session/device.
Used for routing resize requests to the correct R process.
\item \strong{\code{device}} (object):
\itemize{
\item \strong{\code{width}}, \strong{\code{height}}: Device dimensions in pixels.
\item \strong{\code{dpi}}: Dots per inch.
\item \strong{\code{bg}}: Background color as an RGBA string
(see the Color format section).
}
\item \strong{\code{ops}} (array): Drawing operations
(see the Drawing operations section).
}
}

\section{Color format}{


Colors are represented as CSS-style RGBA strings:

\if{html}{\out{<div class="sourceCode">}}\preformatted{"rgba(R,G,B,A)"
}\if{html}{\out{</div>}}
\itemize{
\item R, G, B: integers 0--255.
\item A: decimal 0.0--1.0 (e.g., \code{"rgba(0,0,0,0.502)"}).
}

Transparent or \code{NA} colors are represented as JSON \code{null}.
}

\section{Graphics context}{


Most drawing operations include a \code{"gc"} object (exceptions are
noted per operation):

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{
  "col": "rgba(0,0,0,1)",
  "fill": null,
  "lwd": 1.0,
  "lty": [],
  "lend": "round",
  "ljoin": "round",
  "lmitre": 10.0,
  "font": \{
    "family": "sans",
    "face": 1,
    "size": 12.0,
    "lineheight": 1.2
  \},
  "ext": \{\}
\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{col}}: Stroke color (RGBA string or \code{null}).
\item \strong{\code{fill}}: Fill color (RGBA string or \code{null}).
\item \strong{\code{lwd}}: Line width in pixels (number).
\item \strong{\code{lty}}: Line type as an array of dash lengths (in pixels).
Solid lines and blank (invisible) lines both produce an empty
array \verb{[]}. When the line is blank, \code{col} is \code{null}, so
renderers can distinguish via the color.
\item \strong{\code{lend}}: Line end cap: \code{"round"}, \code{"butt"}, or \code{"square"}.
\item \strong{\code{ljoin}}: Line join: \code{"round"}, \code{"miter"}, or \code{"bevel"}.
\item \strong{\code{lmitre}}: Miter limit (number).
\item \strong{\code{font}}:
\itemize{
\item \strong{\code{family}}: Font family name (string; empty string if
default).
\item \strong{\code{face}}: Font face: 1 = plain, 2 = bold, 3 = italic,
4 = bold italic, 5 = symbol.
\item \strong{\code{size}}: Font size in points (number).
\item \strong{\code{lineheight}}: Line height multiplier (number).
}
\item \strong{\code{ext}} (object, optional): Per-operation extension data.
Present only when set. Free-form JSON.
}
}

\section{Drawing operations}{


Each element of the \code{ops} array is a JSON object with an \code{"op"}
field. Most drawing operations include a \code{"gc"} field (see
the Graphics context section). Exceptions are noted per operation.

All coordinates are in device pixels with a top-left origin (see
the Coordinate system section).

\strong{clip} -- Set the clipping rectangle. No \code{gc}.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "clip", "x0": 0, "y0": 0, "x1": 768, "y1": 576\}
}\if{html}{\out{</div>}}

\strong{line} -- A single line segment.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "line", "x1": 100, "y1": 200,
 "x2": 300, "y2": 400, "gc": \{\}\}
}\if{html}{\out{</div>}}

\strong{polyline} -- Connected line segments (not closed).

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "polyline", "x": [1, 2, 3],
 "y": [4, 5, 6], "gc": \{\}\}
}\if{html}{\out{</div>}}

\strong{polygon} -- Closed polygon (filled and/or stroked).

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "polygon", "x": [1, 2, 3],
 "y": [4, 5, 6], "gc": \{\}\}
}\if{html}{\out{</div>}}

\strong{rect} -- Rectangle.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "rect", "x0": 10, "y0": 20,
 "x1": 100, "y1": 80, "gc": \{\}\}
}\if{html}{\out{</div>}}

\strong{circle} -- Circle.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "circle", "x": 50, "y": 50, "r": 25, "gc": \{\}\}
}\if{html}{\out{</div>}}

\strong{text} -- Text string.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "text", "x": 100, "y": 200, "str": "Hello",
 "rot": 0, "hadj": 0.5, "gc": \{\}\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{str}}: The text content (string).
\item \strong{\code{rot}}: Rotation angle in degrees (counter-clockwise).
\item \strong{\code{hadj}}: Horizontal adjustment (0 = left-aligned,
0.5 = centered, 1 = right-aligned).
}

\strong{path} -- Complex path with subpaths and a fill rule.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "path", "winding": "nonzero",
 "subpaths": [[[10, 20], [30, 40], [50, 20]]],
 "gc": \{\}\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{winding}}: Fill rule: \code{"nonzero"} or \code{"evenodd"}.
\item \strong{\code{subpaths}}: Array of subpaths. Each subpath is an array
of \verb{[x, y]} coordinate pairs.
}

\strong{raster} -- Raster image. No \code{gc}.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "raster", "x": 0, "y": 576, "w": 100, "h": -80,
 "rot": 0, "interpolate": true,
 "pw": 200, "ph": 160,
 "data": "data:image/png;base64,..."\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{x}}, \strong{\code{y}}: Bottom-left corner of the destination
rectangle in device coordinates.
\item \strong{\code{w}}, \strong{\code{h}}: Displayed width and height. May be
\strong{negative} to indicate a horizontal or vertical flip;
renderers should use the absolute value for sizing and adjust
the anchor point accordingly.
\item \strong{\code{pw}}, \strong{\code{ph}}: Pixel width and height of the source
image.
\item \strong{\code{rot}}: Rotation angle in degrees.
\item \strong{\code{interpolate}}: Whether to interpolate when scaling.
\item \strong{\code{data}}: Base64-encoded PNG as a data URI.
}

\strong{beginGroup} -- Start a drawing group (experimental). No \code{gc}.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "beginGroup",
 "ext": \{"filter": "blur(5px)", "opacity": 0.8\}\}
}\if{html}{\out{</div>}}
\itemize{
\item \strong{\code{ext}} (object, optional): Group-level extension data.
Present only when set. Free-form JSON. Common keys: \code{filter}
(CSS filter string), \code{opacity} (number 0--1), \code{blendMode}
(CSS blend mode string).
}

\strong{endGroup} -- End the most recently opened group. No \code{gc}, no
fields other than \code{"op"}.

\if{html}{\out{<div class="sourceCode json">}}\preformatted{\{"op": "endGroup"\}
}\if{html}{\out{</div>}}

Groups nest arbitrarily.
}

\section{Resize protocol}{


The server receives resize messages from the renderer and
forwards them to R. R replays the drawing at the new dimensions
and sends back a frame message.

\strong{Normal resize flow:}

\if{html}{\out{<div class="sourceCode">}}\preformatted{Renderer -> Server:  \{"type":"resize","width":800,
                      "height":600\}
Server   -> R:       \{"type":"resize","width":800,
                      "height":600\}
R        -> Server:  \{"type":"frame",
                      "resizeReplay":true,
                      "incremental":false,...\}
}\if{html}{\out{</div>}}

\strong{History resize flow} (replay a historical plot):

\if{html}{\out{<div class="sourceCode">}}\preformatted{Renderer -> Server:  \{"type":"resize","width":800,
                      "height":600,"plotIndex":2,
                      "sessionId":"r-1234-1"\}
Server   -> R:       \{"type":"resize","width":800,
                      "height":600,"plotIndex":2\}
R        -> Server:  \{"type":"frame",
                      "resizeReplay":true,
                      "plotIndex":2,
                      "incremental":false,...\}
}\if{html}{\out{</div>}}

Note: The server strips \code{sessionId} before forwarding to R.
History resizes are routed only to the R session that owns the
target plot.

\strong{Resize deduplication:}

Servers should deduplicate consecutive normal resizes with
identical dimensions for each R session. However, if the
previous resize was a \code{plotIndex} resize, the next normal resize
at the same dimensions must NOT be deduplicated, because they
target different contexts (historical snapshot vs. current
plot).
}

\section{Multiple R sessions}{


A server may accept connections from multiple R processes
simultaneously. Each R connection has its own \code{sessionId} and
independent state. Servers should:
\itemize{
\item Route \code{metrics_request} messages from each R connection to
the renderer, and route the matching \code{metrics_response} back
to the originating R session. The \code{id} field is only unique
within a single R process, so servers must scope routing by
the originating connection (e.g. keyed by session + \code{id}).
If a server remaps IDs when forwarding to a shared renderer,
it must restore the original \code{id} when relaying the response
back to R.
\item Route \code{plotIndex} resizes to the R session that owns the
target plot (identified by \code{sessionId} in the resize message).
\item Broadcast normal resizes to all connected R sessions.
\item Broadcast \code{frame} and \code{close} messages to all connected
renderers.
}
}

\section{Session ID management}{


The \code{sessionId} in frame messages identifies the R device
instance. Servers should treat it as an \strong{opaque string}. Do
not parse it or make assumptions about its format.

Session ID reuse is unlikely but possible (e.g., after process
restart). As a defensive measure, servers should disambiguate
if a previously retired \code{sessionId} reappears, to prevent
\code{plotIndex} resizes for old plots from reaching the new
connection. The server's (possibly remapped) \code{sessionId} is
what the renderer sees; \code{plotIndex} resizes use it for routing.
}

\section{Connection lifecycle}{


\strong{Graceful close:} The client sends \code{{"type":"close"}} on
device shutdown. The server should forward this to renderers
and clean up routing state for that session.

\strong{Ungraceful disconnect:} If the connection drops without a
\code{close} message (e.g., process crash), the server should detect
the broken connection (EOF or socket error), clean up the
session, and optionally notify renderers.

\strong{Incomplete lines:} If a connection drops mid-line (no
trailing \verb{\\n}), the partial data should be discarded.
}

\section{Extension fields}{


Extension fields (\code{ext}) appear at three levels:
\itemize{
\item \strong{Frame-level}: Top-level \code{ext} on the frame message.
Applies to the entire frame.
\item \strong{Graphics context level}: \code{ext} inside the \code{gc} object.
Applies to all drawing operations while active.
\item \strong{Group level}: \code{ext} on \code{beginGroup} operations. Applies
only to that group.
}

All \code{ext} fields are free-form JSON objects. When unset, the
field is omitted from the message (never sent as \code{null}). When
set, it may contain any JSON object, including an empty \code{{}}.
Servers should preserve and forward them to renderers without
validation. Renderers should ignore unknown keys.

Extension fields are preserved across resize replays, so
historical plot snapshots retain their \code{ext} data.

See \code{\link[=jgd_ext]{jgd_ext()}}, \code{\link[=jgd_frame_ext]{jgd_frame_ext()}}, and \code{\link[=jgd_begin_group]{jgd_begin_group()}}
for the R API to set these fields.
}

\section{Implementing a server}{


A minimal server implementation needs to:
\enumerate{
\item Listen on a Unix socket, named pipe, or TCP port.
\item Accept R connections and read JSONL lines.
\item Send a deferred \code{server_info} welcome after receiving the
first message from R (not before).
\item Forward \code{frame} messages to connected renderers.
\item Forward \code{resize} messages from renderers to R.
\item Handle \code{metrics_request}/\code{metrics_response} routing between
R and the renderer. Clients handle their own timeouts;
servers are not required to synthesize fallback responses.
\item Forward \code{close} messages to renderers and clean up the
session state (remove metrics routing entries, etc.).
}

Optional:
\itemize{
\item Write a discovery file on startup, remove on shutdown.
\item Implement resize deduplication.
\item Implement session ID remapping for reused IDs.
\item Serve an HTTP endpoint for the renderer UI.
}
}

\seealso{
\code{\link[=jgd]{jgd()}}, \code{\link[=jgd_server_info]{jgd_server_info()}}, \code{\link[=jgd_ext]{jgd_ext()}},
\code{\link[=jgd_frame_ext]{jgd_frame_ext()}}, \code{\link[=jgd_begin_group]{jgd_begin_group()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd_begin_group}
\alias{jgd_begin_group}
\title{Begin a drawing group (experimental)}
\usage{
jgd_begin_group(ext = NULL)
}
\arguments{
\item{ext}{A single JSON string with extension fields for this group, or
\code{NULL} for a group without extension fields.}
}
\value{
Called for its side effect; returns \code{NULL} invisibly.
}
\description{
Emits a \code{beginGroup} operation into the drawing stream.  All subsequent
drawing operations until the matching \code{\link[=jgd_end_group]{jgd_end_group()}} are part of this
group.  The renderer may use the group's extension fields to apply effects
to the group as a whole.
}
\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{with_jgd_frame_ext}
\alias{with_jgd_frame_ext}
\title{Scoped frame-level extension fields (experimental)}
\usage{
with_jgd_frame_ext(json, expr)
}
\arguments{
\item{json}{A single JSON string representing the extension object.}

\item{expr}{Expression to evaluate with the frame extension active.}
}
\value{
The result of evaluating \code{expr}.
}
\description{
Temporarily sets frame-level extension fields for the duration of \code{expr},
then clears them on exit.
}
\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd_ext}
\alias{jgd_ext}
\title{Set extended graphics context (experimental)}
\usage{
jgd_ext(json = NULL)
}
\arguments{
\item{json}{A single JSON string representing the extension object, or
\code{NULL} to clear.  The string must be valid JSON (validated on the C side
via cJSON); an error is raised otherwise.  Packages built on top of jgd
(using e.g. jsonlite) should provide user-friendly wrappers.}
}
\value{
Called for its side effect; returns \code{NULL} invisibly.
}
\description{
Sets extension fields that are included in every subsequent drawing
operation's graphics context (\code{gc.ext} in the JSON protocol).  This is
an experimental, low-level API for injecting renderer-specific properties
(e.g. blend modes, shadows, opacity) that go beyond R's standard graphics
parameters.
}
\section{Supported extension fields}{

The Deno reference server and VS Code renderer currently support:\tabular{lll}{
   Field \tab Canvas2D property \tab Example \cr
   \code{blendMode} \tab \code{globalCompositeOperation} \tab \code{"multiply"} \cr
   \code{opacity} \tab \code{globalAlpha} \tab \code{0.5} \cr
   \code{shadow.blur} \tab \code{shadowBlur} \tab \code{10} \cr
   \code{shadow.color} \tab \code{shadowColor} \tab \code{"rgba(0,0,0,0.5)"} \cr
   \code{shadow.offsetX} \tab \code{shadowOffsetX} \tab \code{5} \cr
   \code{shadow.offsetY} \tab \code{shadowOffsetY} \tab \code{5} \cr
   \code{filter} \tab \code{filter} \tab \code{"blur(3px)"} \cr
}


Custom renderers may support additional fields. Unknown fields are
silently ignored, so extensions are forward-compatible.
}

\section{Design for extension packages}{

\code{jgd_ext()} is intentionally low-level — it accepts a raw JSON string.
Higher-level packages built on top of jgd can provide user-friendly
wrappers with proper argument checking, e.g.:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{jgd_shadow = function(blur = 0, color = "black",
                      offsetX = 0, offsetY = 0) \{
  jgd_ext(jsonlite::toJSON(
    list(shadow = list(blur = blur, color = color,
                       offsetX = offsetX, offsetY = offsetY)),
    auto_unbox = TRUE
  ))
\}
}\if{html}{\out{</div>}}

jgd itself has no dependency on jsonlite or any serialization library;
upstream packages choose their own.
}

\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
jgd()

# Drop shadow (scoped -- automatically cleared after the block)
with_jgd_ext(
  '{"shadow":{"blur":15,"color":"rgba(0,0,0,0.5)","offsetX":5,"offsetY":5}}',
  plot(1:10, pch = 19, cex = 3, col = "steelblue")
)

# Semi-transparent overlay
with_jgd_ext('{"opacity":0.3}', {
  plot(1:10, pch = 19, cex = 5, col = "red")
})

# Manual set/clear
jgd_ext('{"blendMode":"multiply"}')
plot(1:10)
jgd_ext(NULL)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=with_jgd_ext]{with_jgd_ext()}}, \code{\link[=jgd_frame_ext]{jgd_frame_ext()}}, \code{\link[=jgd_begin_group]{jgd_begin_group()}},
\link{jgd_spec}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{with_jgd_group}
\alias{with_jgd_group}
\title{Scoped drawing group (experimental)}
\usage{
with_jgd_group(ext, expr)
}
\arguments{
\item{ext}{A single JSON string with extension fields for this group, or
\code{NULL} for a group without extension fields.}

\item{expr}{Expression to evaluate within the group.}
}
\value{
The result of evaluating \code{expr}.
}
\description{
Opens a drawing group with extension fields, evaluates \code{expr}, then closes
the group on exit.
}
\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd_frame_ext}
\alias{jgd_frame_ext}
\title{Set frame-level extension fields (experimental)}
\usage{
jgd_frame_ext(json = NULL)
}
\arguments{
\item{json}{A single JSON string representing the extension object, or
\code{NULL} or \code{""} to clear.  Non-empty strings must be valid JSON; an error
is raised otherwise.}
}
\value{
Called for its side effect; returns \code{NULL} invisibly.
}
\description{
Sets extension fields that are included once per frame in the JSON protocol
(at the top level of the frame message, not per drawing operation).  This is
useful for frame-wide properties such as post-processing effects.
}
\section{Lifecycle}{

\strong{Experimental.} This API may change in future versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd-package.R
\docType{package}
\name{jgd-package}
\alias{jgd-package}
\title{jgd: JSON Graphics Device}
\description{
A graphics device that translates R plotting operations into JSON and streams them over a local connection to an external display application. The device acts as a pure recorder with no rendering dependencies; all rendering occurs in that application (e.g. a 'VS Code' extension or a web browser). Official display applications are available from the project homepage.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/grantmcdermott/jgd}
  \item Report bugs at \url{https://github.com/grantmcdermott/jgd/issues}
}

}
\author{
\strong{Maintainer}: Grant McDermott \email{contact@grantmcdermott.com}

Authors:
\itemize{
  \item Tatsuya Shima \email{ts1s1andn@gmail.com}
}

Other contributors:
\itemize{
  \item Dave Gamble (cJSON library in src/cjson/) [copyright holder]
  \item cJSON contributors (cJSON library in src/cjson/) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jgd.R
\name{jgd_discover}
\alias{jgd_discover}
\title{Discover a running jgd server}
\usage{
jgd_discover()
}
\value{
A named list with \code{server_name} (character), \code{socket_path}
(character), \code{pid} (integer), and \code{server_info} (named character
vector), or \code{NULL} if no discovery file is found.
}
\description{
Reads the jgd discovery file from the platform cache directory
(\verb{~/.cache/jgd} on Linux, \verb{~/Library/Caches/jgd} on macOS,
\verb{\%LOCALAPPDATA\%/jgd} on Windows) and returns its contents.
This does not require an open jgd device — it simply reads the
file that a running server has written.
}
